import magma as m
from magma.generator import Generator, ParamDict
from magma.testing import check_files_equal


def test_add_gen():
    class Add(Generator):
        @staticmethod
        def generate(width):
            """
            Define and return magma circuit
            """
            class AddPrim(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[width]),
                    I1=m.In(m.Bits[width]),
                    O=m.Out(m.Bits[width]),
                )

            class _Add(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[width]),
                    I1=m.In(m.Bits[width]),
                    O=m.Out(m.Bits[width]),
                )
                io.O <= AddPrim()(io.I0, io.I1)
            return _Add

    # Reference the generated circuit definition using generate method
    m.compile("build/AddGen8", Add.generate(8))
    assert check_files_equal(__file__, f"build/AddGen8.v", f"gold/AddGen8.v")

    class Top(m.Circuit):
        io = m.IO(
            I0=m.In(m.Bits[8]),
            I1=m.In(m.Bits[8]),
            O=m.Out(m.Bits[8]),
        )

        # Define and instance circuit by instancing the generator class
        io.O <= Add(8, name='add8')(io.I0, io.I1)

    m.compile("build/AddGen8Top", Top)
    assert check_files_equal(__file__, f"build/AddGen8Top.v",
                             f"gold/AddGen8Top.v")


def test_gen_cache():
    class Add(Generator):
        @staticmethod
        def generate(params: ParamDict):
            class AddPrim(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[params["width"]]),
                    I1=m.In(m.Bits[params["width"]]),
                    O=m.Out(m.Bits[params["width"]]),
                )

            class _Add(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[params["width"]]),
                    I1=m.In(m.Bits[params["width"]]),
                    O=m.Out(m.Bits[params["width"]]),
                )
                io.O <= AddPrim()(io.I0, io.I1)
            return _Add

    assert Add.generate(ParamDict(width=8)) is Add.generate(ParamDict(width=8))

    class AddNoCache(Generator):
        cache = False

        @staticmethod
        def generate(params: ParamDict):
            class AddPrim(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[params["width"]]),
                    I1=m.In(m.Bits[params["width"]]),
                    O=m.Out(m.Bits[params["width"]]),
                )

            class _Add(m.Circuit):
                io = m.IO(
                    I0=m.In(m.Bits[params["width"]]),
                    I1=m.In(m.Bits[params["width"]]),
                    O=m.Out(m.Bits[params["width"]]),
                )

                io.O <= AddPrim()(io.I0, io.I1)
            return _Add

    assert AddNoCache.generate(ParamDict(width=8)) is not \
        AddNoCache.generate(ParamDict(width=8))



import functools
import itertools
from typing import Dict, Optional, Union, Set

from magma.bits import Bits
from magma.circuit import Circuit, CircuitType, CircuitBuilder
from magma.clock import Enable
from magma.conversions import from_bits
from magma.digital import Digital
from magma.primitives.register import AbstractRegister
from magma.primitives.wire import Wire
from magma.ref import DerivedRef
from magma.t import Type, In, Out
from magma.when import (
    BlockBase as WhenBlock,
    no_when,
    find_inferred_latches,
    emit_when_assertions,
)
from magma.value_utils import make_selector
from magma.wire import wire


_ISWHEN_KEY = "_iswhen_"


def iswhen(inst_or_defn: Union[Circuit, CircuitType]) -> bool:
    if isinstance(inst_or_defn, Circuit):
        return getattr(type(inst_or_defn), _ISWHEN_KEY, False)
    return getattr(inst_or_defn, _ISWHEN_KEY, False)


class InferredLatchError(Exception):
    pass


# NOTE(rsetaluri): This is needed because importing magma.primitives.Memory
# results in a circular import.
@functools.lru_cache(maxsize=1)
def _Memory():
    from magma.primitives.memory import Memory
    return Memory


def _is_memory_port(value: Type) -> bool:
    Memory = _Memory()
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return False
    if not isinstance(type(value_inst), Memory):
        return False
    if root_ref.name not in ("WDATA", "WADDR", "WE", "RADDR"):
        return False
    return True


def _add_default_drivers_to_memory_ports(
    builder: 'WhenBuilder', latches: Set[Type]
):
    # TODO(leonardt): for waddr/wdata/raddr, we could use an existing value as a
    # default since ren/wen will be low, this would avoid potentially adding an
    # extra input to the select mux
    for drivee in builder.output_to_index:
        if drivee in builder.default_drivers:
            continue
        if not _is_memory_port(drivee):
            continue
        if drivee not in latches:
            continue
        T = type(drivee).undirected_t
        zero = from_bits(T, Bits[T.flat_length()](0))
        builder.block.add_default_driver(drivee, zero)
        latches.remove(drivee)


def _get_corresponding_register_default(value: Type) -> Optional[Type]:
    """Used to get the corresponding default for `value` if it is a
    reference to a register input value.

    `value` should either be reg.I or reg.CE:
        * For reg.I, the default will be reg.O (default update is the
          current value).
        * For reg.CE, the default will be False (default to no update).

    """
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return None
    value_inst_T = type(value_inst)
    if not isinstance(value_inst_T, AbstractRegister):
        return None
    if value is value_inst_T.get_enable(value_inst):
        return type(value)(False)
    sel = make_selector(value)
    return sel.select(value_inst.O)


def _add_default_drivers_to_register_inputs(
    builder: 'WhenBuilder', latches: Set[Type]
):
    """Adds default drivers to register inputs (reg.O is the default value for
    inputs, False is the default for enables).

    The current implementation assumes the magma internally defined register
    primitive is used, which always has only one input and an optional enable
    port. If we encounter a AbstractRegister, we will drive all inputs with
    reg.O and all enables with reg.I (if they have multiple).

    If we wanted to make this interface more extensible, we could have the user
    provide an API for setting up default values.
    """
    for drivee in builder.output_to_index:
        if drivee not in latches:
            continue
        O = _get_corresponding_register_default(drivee)
        if O is None:
            continue
        builder.block.add_default_driver(drivee, O)
        latches.remove(drivee)


def _rewire_driven_value_or_values(value_or_values, new_value):
    if isinstance(value_or_values, list):
        for x, z in zip(value_or_values, new_value):
            _rewire_driven_value_or_values(x, z)
        return
    value_or_values.unwire(keep_wired_when_contexts=True)
    value_or_values @= new_value


class WhenBuilder(CircuitBuilder):
    def __init__(self, block: WhenBlock):
        super().__init__(name=f"When_{id(block)}")
        self._block = block
        self._condition_counter = itertools.count()
        self._driver_counter = itertools.count()
        self._drivee_counter = itertools.count()
        self._input_counter = itertools.count()
        self._output_counter = itertools.count()
        self._input_to_index = {}
        self._output_to_index = {}
        self._input_to_name = {}
        self._output_to_name = {}
        self._default_drivers = {}
        self._set_definition_attr("primitive", True)
        self._set_definition_attr(_ISWHEN_KEY, True)
        self._set_definition_attr("_builder_", self)
        self._is_when_builder_ = True
        self._when_assert_wires = {}

    @property
    def default_drivers(self) -> Dict[Type, Type]:
        return self._default_drivers.copy()

    @property
    def input_to_index(self) -> Dict[Type, int]:
        return self._input_to_index.copy()

    @property
    def output_to_index(self) -> Dict[Type, int]:
        return self._output_to_index.copy()

    @property
    def output_to_name(self):
        return self._output_to_name

    @property
    def input_to_name(self):
        return self._input_to_name

    def check_existing_derived_ref(self, value, value_to_name, value_to_index):
        """If value is a child of an array or tuple that has already been added,
        we return the child of the existing value, rather than adding a new
        port, which allows us to maintain bulk assignments in the eventual
        generated if statement, rather that elaborating into per-child
        assignments.
        """
        curr_root = None
        for ref in value.name.root_iter(
            stop_if=lambda ref: not isinstance(ref, DerivedRef)
        ):
            if ref.parent_value in value_to_index:
                curr_root = ref.parent_value
        if not curr_root:
            return
        root_port = getattr(self, value_to_name[curr_root])
        return make_selector(value, stop_at=curr_root).select(root_port)

    def _generic_add(
            self,
            value,
            value_to_index,
            index_counter,
            name_counter,
            value_to_name,
            name_prefix,
            type_qualifier,
    ):
        if value in value_to_index:
            return
        value_to_index[value] = next(index_counter)
        port_name = f"{name_prefix}{next(name_counter)}"
        value_to_name[value] = port_name

        port = self.check_existing_derived_ref(value, value_to_name,
                                               value_to_index)
        # NOTE(leonardt): when we add support for flatten_all_tuples=False, we
        # should also add similar logic here to avoid flattening assignments
        if port is None:
            self._add_port(port_name, type_qualifier(type(value).undirected_t))
            port = getattr(self, port_name)
            port.is_when_port = True

        with no_when():
            if value.is_input() and isinstance(value, Enable):
                value.wire(
                    port,
                    driven_implicitly_by_when=value.driven_implicitly_by_when
                )
            elif port.is_input() and port.driven():
                port.rewire(value)
            else:
                wire(port, value)
                port.debug_info = self.debug_info
                value.debug_info = self.debug_info

    def add_drivee(self, value: Type):
        self._generic_add(
            value,
            self._output_to_index,
            self._output_counter,
            self._drivee_counter,
            self._output_to_name,
            "O",
            Out,
        )

    def add_driver(self, value: Type):
        self._generic_add(
            value,
            self._input_to_index,
            self._input_counter,
            self._driver_counter,
            self._input_to_name,
            "I",
            In,
        )

    def add_condition(self, condition: Optional[Digital]):
        self._generic_add(
            condition,
            self._input_to_index,
            self._input_counter,
            self._condition_counter,
            self._input_to_name,
            "S",
            In,
        )

    def add_default_driver(self, drivee: Type, driver: Type):
        self.add_driver(driver)
        self.add_drivee(drivee)
        self._default_drivers[drivee] = driver

    def remove_default_driver(self, drivee: Type):
        del self._default_drivers[drivee]

    def infer_latches(self):
        """
        Detect latches which would be inferred from the context of the when
        block.
        """
        latches = find_inferred_latches(self.block)
        # NOTE(rsetaluri): These passes should ideally be done after circuit
        # creation. However, it is quite unwieldy, so we opt to do it in this
        # builder's finalization, although it is a bit "hacky". Ideally, this
        # builder should be completely agnostic of anything it is connected to,
        # and a global (post-processing) pass should be used instead.
        # NOTE(leonardt): We only add default drivers if a value is inferred to
        # be a latch, otherwise we skip this implicit logic so the output is
        # cleaner (e.g. if a register is assigned in every case, we don't need
        # the default value).
        _add_default_drivers_to_memory_ports(self, latches)
        _add_default_drivers_to_register_inputs(self, latches)

        if latches:
            raise InferredLatchError(latches)

    def emit_when_assertions(self, flatten_all_tuples: bool):
        """
        We run this step before finalize since emitting when asserts may cause
        tuple elaboration (since verilog assert references may refer to a
        tuple that will eventually be flattened). This may cause changes to the
        builder and any connected builders so we don't want to finalize when
        builders before this logic is finished across all instances.

        Must be done after implicit default driver logic is added.
        """
        emit_when_assertions(self.block, self, flatten_all_tuples)

    def _finalize(self):
        pass

    @property
    def block(self) -> WhenBlock:
        return self._block

    def get_when_assert_wire(self, port):
        if port not in self._when_assert_wires:
            name = f"_WHEN_ASSERT_{len(self._when_assert_wires)}"
            # NOTE(leonardt): Here we use the `Wire` module instead of the
            # named value logic because insert_coreir_wires can cause
            # elaboration which then causes a bad interaction with the when
            # finalization logic. Since whens are currently being finalized,
            # the builders cannot be modified later.
            temp = Wire(type(port).undirected_t, flatten=False)(name=name)
            for value in port.driving():
                _rewire_driven_value_or_values(value, temp.O)
            temp.I @= port
            self._when_assert_wires[port] = temp.O
        return self._when_assert_wires[port]


def is_when_builder(builder):
    return isinstance(builder, WhenBuilder)



import pytest

import magma as m
import magma.testing


def _parameterize_by_backend(fn):
    return pytest.mark.parametrize("backend", ("coreir-verilog", "mlir"))(fn)


def _run_test(dut, bound_module, basename, backend):
    m.compile(f"build/{basename}_{backend}", dut, output=backend)

    suffix = "mlir" if backend == "mlir" else "v"
    assert m.testing.check_files_equal(
        __file__,
        f"build/{basename}_{backend}.{suffix}",
        f"gold/{basename}_{backend}.{suffix}")
    assert m.testing.check_files_equal(
        __file__,
        f"build/{bound_module.name}.sv",
        f"gold/{bound_module.name}.sv")
    list_filename = f"tests/build/{basename}_{backend}_bind_files.list"
    with open(list_filename, "r") as f:
        list_contents = f.read()
    assert list_contents == f"{bound_module.name}.sv"


@_parameterize_by_backend
def test_basic(backend):

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= io.I

    class TopBasicAsserts(m.Circuit):
        name = f"TopBasicAsserts_{backend}"
        io = m.IO(I=m.In(m.Bit), O=m.In(m.Bit), other=m.In(m.Bit))
        io.I.unused()
        io.O.unused()
        io.other.unused()

    Top.bind(TopBasicAsserts, Top.I)

    _run_test(Top, TopBasicAsserts, "test_bind_basic", backend)


@_parameterize_by_backend
def test_xmr(backend):

    class Bottom(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= io.I

    class Middle(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= Bottom(name="bottom")(io.I)

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        middle = Middle(name="middle")
        io.O @= middle(io.I)

    class TopXMRAsserts(m.Circuit):
        name = f"TopXMRAsserts_{backend}"
        io = m.IO(I=m.In(m.Bit), O=m.In(m.Bit), other=m.In(m.Bit))
        io.I.unused()
        io.O.unused()
        io.other.unused()

    Top.bind(TopXMRAsserts, Top.middle.bottom.I)

    _run_test(Top, TopXMRAsserts, "test_bind_xmr", backend)



from typing import Iterable, Mapping, Tuple

from ast_tools.stack import get_symbol_table

from magma.circuit import Circuit
from magma.common import hash_expr
from magma.generator import Generator2
from magma.interface import IO
from magma.t import In, Kind, Type
from magma.view import PortView
from magma.wire_utils import wire_value_or_port_view, WiringError
# TODO(rsetaluri): Make these common.
from magma.inline_verilog import (
    _process_fstring_syntax,
    _make_inline_value,
    ValueLike,
    InlineVerilogError,
)


ValueLikeMap = Mapping[str, ValueLike]


def _process_expr(
        expr: str,
        format_args: ValueLikeMap,
) -> Tuple[str, ValueLikeMap]:
    symbol_table = get_symbol_table([inline_verilog2], copy_locals=True)
    value_map = {}
    expr = _process_fstring_syntax(
        expr, format_args, value_map, symbol_table
    )
    # TODO(rsetaluri): Reuse this from magma/inline_verilog.py.
    format_args = format_args.copy()
    for key, arg in format_args.items():
        if isinstance(arg, (Type, PortView)):
            arg = _make_inline_value(value_map, arg)
        format_args[key] = arg
    expr = expr.format(**format_args)
    # Replace all __magma_inline_value_x with x. Note that the value_map is
    # assumed to be ordered, specifically, in the order of the arguments. We
    # need to iterate in reverse fashion to avoid clobbering later keys,
    # e.g. 's/__magma_inline_value_1/{1}/g' would convert
    # '__magma_inline_value_14' to '14'. It is sufficient to iterate in reverse
    # order.
    for i, key in reversed(list(enumerate(value_map.keys()))):
        expr = expr.replace(key, f"{{{str(i)}}}")
    return expr, value_map


def _get_arg_type(value: ValueLike) -> Kind:
    if isinstance(value, PortView):
        return value.T
    return type(value)


def _wire_ports(value_map: ValueLikeMap, inst: Circuit):
    for i, value in enumerate(value_map.values()):
        try:
            wire_value_or_port_view(getattr(inst, f"I{i}"), value)
        except WiringError:
            raise InlineVerilogError(
                f"Found reference to undriven input port: {repr(value)}"
            ) from None


class _InlineVerilog2(Generator2):
    def __init__(self, expr: str, arg_types: Iterable[Kind]):
        self.expr = expr
        self.io = IO(**{
            f"I{i}": In(T)
            for i, T in enumerate(arg_types)
        })
        self.primitive = True
        self.name = f"InlineVerilog2_{hash_expr(expr)}"


InlineVerilog2 = _InlineVerilog2


def inline_verilog2(expr, **kwargs):
    expr, value_map = _process_expr(expr, kwargs)
    inst = _InlineVerilog2(expr, map(_get_arg_type, value_map.values()))()
    _wire_ports(value_map, inst)



import pytest
from typing import Optional

from magma.symbol_table import (SymbolTable, SYMBOL_TABLE_EMPTY,
                                SYMBOL_TABLE_INLINED_INSTANCE)
from magma.symbol_table_utils import (make_master_symbol_table, MasterSymbolTable,
                                      SymbolQueryInterface)


def _make_empty(name: str):
    return (SYMBOL_TABLE_EMPTY, name)


def _make_inlined(key: Optional[str] = None):
    name = "" if key is None else key
    return (SYMBOL_TABLE_INLINED_INSTANCE, name)


def test_one():
    master = MasterSymbolTable([
        SymbolTable.from_dict({
            "module_names": {
                "Top": "CTop",
                "MyModule": "MyModule",
            },
            "instance_names": {
                ("Top", "inst"): _make_empty("c_inst"),
            },
            "port_names": {
                ("Top", "I"): "in",
            },
            "instance_types": {
                ("Top", "inst"): "MyModule",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "CTop": "VTop",
                "MyModule": "MyModule",
            },
            "instance_names": {
                ("CTop", "c_inst"): _make_empty("v_inst"),
            },
            "port_names": {
                ("CTop", "in.x"): "in_x",
                ("CTop", "in.y"): "in_y",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "VTop": "OTop",
                "MyModule": "MyModule",
            },
            "instance_names": {
                ("VTop", "v_inst"): _make_empty("o_inst"),
            },
            "port_names": {
                ("VTop", "in_x"): "in_x",
                ("VTop", "in_y"): "in_y",
            },
        }),
    ])
    dct = master.as_dict()
    assert dct == {
        "module_names": {
            "Top": "OTop",
            "MyModule": "MyModule",
        },
        "instance_names": {
            ("Top", "inst"): _make_empty("o_inst"),
        },
        "port_names": {
            ("Top", "I.x"): "in_x",
            ("Top", "I.y"): "in_y",
        },
        "inlined_instance_names": {},
        "instance_types": {
            ("Top", "inst"): "MyModule",
        },
    }
    tli = SymbolQueryInterface(master)
    assert tli.get_module_name("Top") == "OTop"
    assert tli.get_instance_name("Top.inst") == "OTop.o_inst"


def test_two():
    master = MasterSymbolTable([
        SymbolTable.from_dict({
            "module_names": {
                "x": "y",
                "a": "b",
                "MyModule": "MyModule",
                "Foo": "Foo",
            },
            "instance_names": {
                ("x", "inst0"): _make_empty("inst_0"),
                ("a", "foo"): _make_empty("foo_out"),
            },
            "instance_types": {
                ("x", "inst0"): "MyModule",
                ("a", "foo"): "Foo",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "y": "z",
                "b": "b",
                "MyModule": "MyModule",
                "Foo": "Foo",
            },
            "instance_names": {
                ("y", "inst_0"): _make_empty("inst_2"),
                ("b", "foo_out"): _make_empty("foo_out"),
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "z": "z",
                "b": "f",
                "MyModule": "MyModule",
                "Foo": "Foo",
            },
            "instance_names": {
                ("z", "inst_2"): _make_empty("inst_100"),
                ("b", "foo_out"): _make_empty("foo_OUT"),
            },
        }),
    ])
    dct = master.as_dict()
    assert dct == {
        "module_names": {
            "x": "z",
            "a": "f",
            "MyModule": "MyModule",
            "Foo": "Foo",
        },
        "instance_names": {
            ("x", "inst0"): _make_empty("inst_100"),
            ("a", "foo"): _make_empty("foo_OUT")
        },
        "port_names": {},
        "inlined_instance_names": {},
        "instance_types": {
            ("x", "inst0"): "MyModule",
            ("a", "foo"): "Foo"
        },
    }


def test_inlining_simple():
    master = MasterSymbolTable([
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
            },
            "instance_names": {
                ("M1", "i2"): _make_empty("i2"),
                ("M2", "i3"): _make_empty("i3"),
            },
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
            },
            "instance_names": {
                ("M1", "i2"): _make_inlined(),
                ("M2", "i3"): _make_empty("i3"),
            },
            "inlined_instance_names": {
                ("M1", "i2", "i3"): _make_empty("i2$i3"),
            },
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
            },
        }),
    ])
    dct = master.as_dict()
    assert dct == {
        "module_names": {
            "M1": "M1",
            "M2": "M2",
            "M3": "M3",
        },
        "instance_names": {
            ("M1", "i2"): _make_inlined(),
            ("M2", "i3"): _make_empty("i3"),
        },
        "port_names": {},
        "inlined_instance_names": {
            ("M1", "i2", "i3"): _make_empty("i2$i3"),
        },
        "instance_types": {
            ("M1", "i2"): "M2",
            ("M2", "i3"): "M3",
        },
    }
    tli = SymbolQueryInterface(master)
    assert tli.get_instance_name("M1.i2.i3") == "M1.i2$i3"
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M1.i2")


def test_inlining_top_down():
    master = MasterSymbolTable([
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
                "M4": "M4",
                "M5": "M5",
                "M6": "M6",
            },
            "instance_names": {
                ("M1", "i2"): _make_empty("i2"),
                ("M2", "i3"): _make_empty("i3"),
                ("M3", "i4"): _make_empty("i4"),
                ("M4", "i5"): _make_empty("i5"),
                ("M5", "i6"): _make_empty("i6"),
            },
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
                ("M3", "i4"): "M4",
                ("M4", "i5"): "M5",
                ("M5", "i6"): "M6",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
                "M4": "M4",
                "M5": "M5",
                "M6": "M6",
            },
            "instance_names": {
                ("M1", "i2"): _make_inlined(),
                ("M2", "i3"): _make_empty("i3"),
                ("M3", "i4"): _make_empty("i4"),
                ("M4", "i5"): _make_empty("i5"),
                ("M5", "i6"): _make_empty("i6"),
            },
            "inlined_instance_names": {
                ("M1", "i2", "i3"): _make_inlined("UNIQ_KEY_0"),
                ("M1", "UNIQ_KEY_0", "i4"): _make_empty("i2$i3$i4"),
            },
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
                ("M3", "i4"): "M4",
                ("M4", "i5"): "M5",
                ("M5", "i6"): "M6",
            },
        }),
    ])
    dct = master.as_dict()
    assert dct == {
        "module_names": {
            "M1": "M1",
            "M2": "M2",
            "M3": "M3",
            "M4": "M4",
            "M5": "M5",
            "M6": "M6",
        },
        "instance_names": {
            ("M1", "i2"): _make_inlined(),
            ("M2", "i3"): _make_empty("i3"),
            ("M3", "i4"): _make_empty("i4"),
            ("M4", "i5"): _make_empty("i5"),
            ("M5", "i6"): _make_empty("i6"),
        },
        "port_names": {},
        "inlined_instance_names": {
            ("M1", "i2", "i3"): _make_inlined("UNIQ_KEY_0"),
            ("M1", "UNIQ_KEY_0", "i4"): _make_empty("i2$i3$i4"),
        },
        "instance_types": {
            ("M1", "i2"): "M2",
            ("M2", "i3"): "M3",
            ("M3", "i4"): "M4",
            ("M4", "i5"): "M5",
            ("M5", "i6"): "M6",
        },
    }
    tli = SymbolQueryInterface(master)
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M1.i2")
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M1.i2.i3")
    assert tli.get_instance_name("M1.i2.i3.i4") == "M1.i2$i3$i4"
    assert tli.get_instance_name("M1.i2.i3.i4.i5.i6") == "M1.i2$i3$i4.i5.i6"
    assert tli.get_instance_name("M2.i3") == "M2.i3"
    assert tli.get_instance_name("M2.i3.i4") == "M2.i3.i4"


def test_inlining_bottom_up():
    master = MasterSymbolTable([
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
                "M4": "M4",
                "M5": "M5",
                "M6": "M6",
            },
            "instance_names": {
                ("M1", "i2"): _make_empty("i2"),
                ("M2", "i3"): _make_empty("i3"),
                ("M3", "i4"): _make_empty("i4"),
                ("M4", "i5"): _make_empty("i5"),
                ("M5", "i6"): _make_empty("i6"),
            },
            "port_names": {},
            "inlined_instance_names": {},
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
                ("M3", "i4"): "M4",
                ("M4", "i5"): "M5",
                ("M5", "i6"): "M6",
            },
        }),
        SymbolTable.from_dict({
            "module_names": {
                "M1": "M1",
                "M2": "M2",
                "M3": "M3",
                "M4": "M4",
                "M5": "M5",
                "M6": "M6",
            },
            "instance_names": {
                ("M1", "i2"): _make_inlined(),
                ("M2", "i3"): _make_inlined(),
                ("M3", "i4"): _make_empty("i4"),
                ("M4", "i5"): _make_empty("i5"),
                ("M5", "i6"): _make_empty("i6"),
            },
            "port_names": {},
            "inlined_instance_names": {
                ("M1", "i2", "i3"): _make_inlined("UNIQ_KEY_0"),
                ("M1", "UNIQ_KEY_0", "i4"): _make_empty("i2$i3$i4"),
                ("M2", "i3", "i4"): _make_empty("i3$i4"),
            },
            "instance_types": {
                ("M1", "i2"): "M2",
                ("M2", "i3"): "M3",
                ("M3", "i4"): "M4",
                ("M4", "i5"): "M5",
                ("M5", "i6"): "M6",
            },
        }),
    ])
    dct = master.as_dict()
    assert dct == {
        "module_names": {
            "M1": "M1",
            "M2": "M2",
            "M3": "M3",
            "M4": "M4",
            "M5": "M5",
            "M6": "M6",
        },
        "instance_names": {
            ("M1", "i2"): _make_inlined(),
            ("M2", "i3"): _make_inlined(),
            ("M3", "i4"): _make_empty("i4"),
            ("M4", "i5"): _make_empty("i5"),
            ("M5", "i6"): _make_empty("i6"),
        },
        "port_names": {},
        "inlined_instance_names": {
            ("M1", "i2", "i3"): _make_inlined("UNIQ_KEY_0"),
            ("M1", "UNIQ_KEY_0", "i4"): _make_empty("i2$i3$i4"),
            ("M2", "i3", "i4"): _make_empty("i3$i4"),
        },
        "instance_types": {
            ("M1", "i2"): "M2",
            ("M2", "i3"): "M3",
            ("M3", "i4"): "M4",
            ("M4", "i5"): "M5",
            ("M5", "i6"): "M6",
        },
    }
    tli = SymbolQueryInterface(master)
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M1.i2")
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M1.i2.i3")
    assert tli.get_instance_name("M1.i2.i3.i4") == "M1.i2$i3$i4"
    assert tli.get_instance_name("M1.i2.i3.i4.i5.i6") == "M1.i2$i3$i4.i5.i6"
    with pytest.raises(SymbolQueryInterface.InlinedLeafInstanceError):
        tli.get_instance_name("M2.i3")
    assert tli.get_instance_name("M2.i3.i4") == "M2.i3$i4"
    assert tli.get_instance_name("M2.i3.i4.i5") == "M2.i3$i4.i5"



import functools
import pytest
import tempfile

import magma as m
from magma.testing import check_files_equal


class _BinOpInterface(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))


class _OrImpl(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))
    io.O @= io.I0 | io.I1


class _AndImpl(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))
    io.O @= io.I0 & io.I1


class _XorImpl(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))
    io.O @= io.I0 ^ io.I1


class _Top(m.Circuit):
    io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit))
    io.O @= _BinOpInterface()(io.I0, io.I1)


def _wrap_with_clear_link_info(fn):

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        m.clear_link_info()
        return fn(*args, **kwargs)

    return wrapped


def _compile(output, basename=None):
    if output == "coreir":
        with tempfile.TemporaryDirectory() as tempdir:
            basename = f"{tempdir}/Top"
            m.compile(basename, _Top, output="coreir")
        modules = m.backend.coreir.coreir_runtime.module_map()
        return modules[list(modules.keys())[0]]['global']
    if output == "mlir":
        m.compile(f"build/{basename}", _Top, output="mlir")
        return
    raise NotImplementedError(output)


def _make_basename(output, basename):
    if output == "coreir":
        return None
    if output == "mlir":
        return basename
    raise NotImplementedError(output)


def _run_file_check(basename):
    assert check_files_equal(
        __file__, f"build/{basename}.mlir", f"gold/{basename}.mlir")


@_wrap_with_clear_link_info
@pytest.mark.parametrize("output", ("coreir", "mlir"))
def test_only_default(output):
    m.link_default_module(_BinOpInterface, _OrImpl)
    assert m.linking.has_default_linked_module(_BinOpInterface)
    assert m.linking.get_default_linked_module(_BinOpInterface) is _OrImpl
    basename = _make_basename(output, "test_module_linking_only_default")
    modules = _compile(output, basename=basename)
    if output == "coreir":
        linked = modules["_BinOpInterface"].get_default_linked_module()
        assert linked is modules["_OrImpl"]
        assert not modules["_BinOpInterface"].get_linked_modules()  # expect empty
        return
    if output == "mlir":
        _run_file_check(basename)
        return
    raise NotImplementedError(output)


@_wrap_with_clear_link_info
@pytest.mark.parametrize("output", ("coreir", "mlir"))
def test_linked_modules_no_default(output):
    m.link_module(_BinOpInterface, "OR", _OrImpl)
    m.link_module(_BinOpInterface, "AND", _AndImpl)
    assert not m.linking.has_default_linked_module(_BinOpInterface)
    assert m.linking.get_linked_modules(_BinOpInterface) == {
        "OR": _OrImpl,
        "AND": _AndImpl,
    }
    basename = _make_basename(output, "test_module_linking_no_default")
    modules = _compile(output, basename=basename)
    if output == "coreir":
        assert not modules["_BinOpInterface"].has_default_linked_module()
        linked = modules["_BinOpInterface"].get_linked_modules()
        assert len(linked) == 2
        assert linked["OR"] is modules["_OrImpl"]
        assert linked["AND"] is modules["_AndImpl"]
        return
    if output == "mlir":
        _run_file_check(basename)
        return
    raise NotImplementedError(output)


@_wrap_with_clear_link_info
@pytest.mark.parametrize("output", ("coreir", "mlir"))
def test_linked_modules_with_default(output):
    m.link_module(_BinOpInterface, "OR", _OrImpl)
    m.link_module(_BinOpInterface, "AND", _AndImpl)
    m.link_default_module(_BinOpInterface, _OrImpl)
    assert m.linking.get_linked_modules(_BinOpInterface) == {
        "OR": _OrImpl,
        "AND": _AndImpl,
    }
    assert m.linking.has_default_linked_module(_BinOpInterface)
    assert m.linking.get_default_linked_module(_BinOpInterface) is _OrImpl
    basename = _make_basename(output, "test_module_linking_with_default")
    modules = _compile(output, basename=basename)
    if output == "coreir":
        linked = modules["_BinOpInterface"].get_default_linked_module()
        assert linked is modules["_OrImpl"]
        linked = modules["_BinOpInterface"].get_linked_modules()
        assert len(linked) == 2
        assert linked["OR"] is modules["_OrImpl"]
        assert linked["AND"] is modules["_AndImpl"]
        return
    if output == "mlir":
        _run_file_check(basename)
        return
    raise NotImplementedError(output)



from magma import *

def test_flatten():
    b0 = Bit(name='b0')
    b1 = Bit(name='b1')
    b2 = Bit(name='b2')
    b3 = Bit(name='b3')
    b4 = Bit(name='b4')
    a1 = array([b0, b1])
    a2 = array([b0, b1])
    print(a1==a2)
    A = Array[2,Bit]
    B = Array[2,Bit]
    print(A==B)
    t = tuple_(dict(x=array([b0, b1]), y=array([b2, b3, b4])))
    print(type(t.name))
    print(type(t))
    a = array(t.flatten())
    print(a.name)
    print(type(a))



import dataclasses
import pathlib

from magma.backend.mlir.mlir import (
    begin_dialect,
    end_dialect,
    MlirAttribute,
    MlirBlock,
    MlirDialect,
    MlirLocation,
    MlirOp,
    MlirRegion,
    MlirType,
)
from magma.backend.mlir.mlir_printer_utils import print_attr_dict
from magma.backend.mlir.print_opts import PrintOpts
from magma.backend.mlir.printer_base import PrinterBase
from magma.config import config, RuntimeConfig


# TODO(rsetaluri): This global runtime config is really somewhat of a hack. This
# should really be a pass or a compile-time option.
config.register(mlir_emit_loc_filename_only=RuntimeConfig(False))

builtin = MlirDialect("builtin")
begin_dialect(builtin)


class UnknownLoc(MlirLocation):
    def emit(self) -> str:
        return "loc(unknown)"


@dataclasses.dataclass(frozen=True)
class FileLineColLoc(MlirLocation):
    file: str
    line: int
    col: int

    def emit(self) -> str:
        path = pathlib.Path(self.file)
        if config.mlir_emit_loc_filename_only:
            path = path.name
        return f"loc(\"{path}\":{self.line}:{self.col})"


@dataclasses.dataclass(frozen=True)
class IntegerType(MlirType):
    n: int

    def emit(self) -> str:
        return f"i{self.n}"


@dataclasses.dataclass(frozen=True)
class IntegerAttr(MlirAttribute):
    value: int

    def emit(self) -> str:
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class BoolAttr(MlirAttribute):
    value: bool

    def emit(self) -> str:
        return "true" if self.value else "false"


@dataclasses.dataclass(frozen=True)
class StringAttr(MlirAttribute):
    value: str

    def emit(self) -> str:
        return f"\"{self.value}\""


@dataclasses.dataclass
class ModuleOp(MlirOp):
    def __post_init__(self):
        self._block = self.new_region().new_block()

    @property
    def block(self) -> MlirBlock:
        return self._block

    def add_operation(self, operation: MlirOp):
        self._block.add_operation(operation)

    def print_op(self, printer: PrinterBase, print_opts: PrintOpts):
        printer.print("module")
        if self.attr_dict:
            printer.print(" attributes ")
            print_attr_dict(self.attr_dict, printer)


end_dialect()



import pytest
import magma as m
from magma.config import config
from magma.testing.utils import check_gold


@pytest.fixture(autouse=True)
def use_namer_dict():
    config.use_namer_dict = True
    yield
    config.use_namer_dict = False


def test_basic_namer_dict():

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))

    class test_basic_namer_dict(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
        foo = Foo()
        assert foo.name == "foo"
        x = m.Bits[8]()
        assert x.name.name == "x"
        x @= io.I
        io.O @= foo(x)

    m.compile("build/test_basic_namer_dict", test_basic_namer_dict,
              output="mlir")
    assert check_gold(__file__, f"test_basic_namer_dict.mlir")


def test_namer_dict_multiple():

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))

    class test_namer_dict_multiple(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
        foo = Foo()
        x = m.Bits[8]()
        x @= io.I
        y = foo(x)
        foo = Foo()
        x = m.Bits[8]()
        x @= y
        io.O @= foo(x)

    m.compile("build/test_namer_dict_multiple", test_namer_dict_multiple,
              output="mlir")
    assert check_gold(__file__, f"test_namer_dict_multiple.mlir")


def test_namer_dict_smart_bits():

    class test_namer_dict_smart_bits(m.Circuit):
        io = m.IO(I0=m.In(m.smart.SmartBits[8]),
                  I1=m.In(m.smart.SmartBits[8]),
                  O=m.Out(m.smart.SmartBits[9]))
        x = m.smart.SmartBits[9]()
        x @= io.I0 + io.I1
        io.O @= x

    m.compile("build/test_namer_dict_smart_bits", test_namer_dict_smart_bits,
              output="mlir")
    assert check_gold(__file__, f"test_namer_dict_smart_bits.mlir")


def test_namer_dict_smart_bits_lazy():

    class test_namer_dict_smart_bits_lazy(m.Circuit):
        io = m.IO(I0=m.In(m.smart.SmartBits[8]),
                  I1=m.In(m.smart.SmartBits[8]),
                  O=m.Out(m.smart.SmartBits[9]))
        x = io.I0 + io.I1
        io.O @= x

    m.compile("build/test_namer_dict_smart_bits_lazy",
              test_namer_dict_smart_bits_lazy,
              output="mlir")
    assert check_gold(__file__, f"test_namer_dict_smart_bits_lazy.mlir")


def test_namer_dict_smart_bits_lazy_rename():

    class test_namer_dict_smart_bits_lazy_rename(m.Circuit):
        io = m.IO(I0=m.In(m.smart.SmartBits[8]),
                  I1=m.In(m.smart.SmartBits[8]),
                  O=m.Out(m.smart.SmartBits[9]))
        x = io.I0 + io.I1
        y = x
        io.O @= y

    m.compile("build/test_namer_dict_smart_bits_lazy_rename",
              test_namer_dict_smart_bits_lazy_rename,
              output="mlir")
    assert check_gold(__file__, f"test_namer_dict_smart_bits_lazy_rename.mlir")


def test_namer_dict_already_named():

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))

    class test_namer_dict_already_named(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
        foo = Foo(name="bar")
        assert foo.name == "bar"
        x = m.Bits[8](name="y")
        assert x.name.name == "y"
        x @= io.I
        io.O @= foo(x)

    m.compile("build/test_namer_dict_already_named",
              test_namer_dict_already_named, output="mlir")
    assert check_gold(__file__, f"test_namer_dict_already_named.mlir")


def test_namer_dict_generator():

    class Foo(m.Generator2):
        def __init__(self, n):
            self.io = m.IO(I=m.In(m.Bits[n]), O=m.Out(m.Bits[n]))
            self.x = m.Bits[n]()
            self.x @= self.io.I
            self.io.O @= self.x

    m.compile("build/test_namer_dict_generator", Foo(8), output="mlir")
    assert check_gold(__file__, f"test_namer_dict_generator.mlir")


def test_namer_dict_explicit_collision_inst():

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))

    class test_namer_dict_explicit_collision_inst(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
        x = Foo(name="foo")
        foo = Foo()
        io.O @= foo(x(io.I))

    m.compile("build/test_namer_dict_explicit_collision_inst",
              test_namer_dict_explicit_collision_inst, output="mlir")
    assert check_gold(__file__, f"test_namer_dict_explicit_collision_inst.mlir")


def test_namer_dict_explicit_collision_value():

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))

    class test_namer_dict_explicit_collision_value(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
        x = m.Bits[8](name="y")
        x @= io.I
        y = m.Bits[8]()
        y @= x
        io.O @= y

    m.compile("build/test_namer_dict_explicit_collision_value",
              test_namer_dict_explicit_collision_value, output="mlir")
    assert check_gold(__file__,
                      f"test_namer_dict_explicit_collision_value.mlir")



import pytest
import random
import string

from magma.symbol_table import (SymbolTable, SYMBOL_TABLE_INLINED_INSTANCE,
                                SYMBOL_TABLE_EMPTY)


_NUM_TRIALS = 100
_NAME_SIZE = 10


def _random_name(size):
    return ''.join(random.choice(string.ascii_letters) for _ in range(size))


def _check_equal(l, r):
    if isinstance(l, tuple) and not isinstance(r, tuple):
        r = (r,)
    elif isinstance(r, tuple) and not isinstance(l, tuple):
        l = (l,)
    return l == r


def _test_basic_mapping(setter, getter, in_size, out_size, num_trials):
    golden = {}
    for i in range(num_trials):
        ins = tuple(_random_name(_NAME_SIZE) for _ in range(in_size))
        outs = tuple(_random_name(_NAME_SIZE) for _ in range(out_size))
        expect_error = ins in golden
        if not expect_error:
            setter(*(ins + outs))
            _check_equal(getter(*ins), outs)
            continue
        with pytest.raises(KeyError) as pytest_e:
            setter(*ins)
            assert False
        assert pytest_e.type is KeyError


def test_module_name():
    table = SymbolTable()

    def _setter(*args):
        table.set_module_name(*args)

    def _getter(*args):
        return table.get_module_name(*args)

    _test_basic_mapping(_setter, _getter, 1, 1, _NUM_TRIALS)


def test_instance_name():
    # NOTE(rsetaluri): Right now this test does not inject values of the correct
    # type, but since the symbol table API does not check the type of values for
    # the in memory representation, it doesn't matter. This should be fixed.
    table = SymbolTable()

    def _setter(*args):
        table.set_instance_name(*args)

    def _getter(*args):
        return table.get_instance_name(*args)

    _test_basic_mapping(_setter, _getter, 2, 1, _NUM_TRIALS)


def test_port_name():
    table = SymbolTable()

    def _setter(*args):
        table.set_port_name(*args)

    def _getter(*args):
        return table.get_port_name(*args)

    _test_basic_mapping(_setter, _getter, 2, 1, _NUM_TRIALS)


def test_inlined_instance_name_name():
    # NOTE(rsetaluri): Right now this test does not inject values of the correct
    # type, but since the symbol table API does not check the type of values for
    # the in memory representation, it doesn't matter. This should be fixed.
    table = SymbolTable()

    def _setter(*args):
        table.set_inlined_instance_name(*args)

    def _getter(*args):
        return table.get_inlined_instance_name(*args)

    _test_basic_mapping(_setter, _getter, 3, 1, _NUM_TRIALS)


def test_instance_type():
    table = SymbolTable()

    def _setter(*args):
        table.set_instance_type(*args)

    def _getter(*args):
        return table.get_instance_type(*args)

    _test_basic_mapping(_setter, _getter, 2, 1, _NUM_TRIALS)


def test_inline():
    table = SymbolTable()
    value = (SYMBOL_TABLE_INLINED_INSTANCE, "")
    table.set_instance_name("Foo", "bar", value)
    assert table.get_instance_name("Foo", "bar") == value
    value = (SYMBOL_TABLE_EMPTY, "bar_x")
    table.set_inlined_instance_name("Foo", "bar", "x", value)
    assert table.get_inlined_instance_name("Foo", "bar", "x") == value


def test_json():
    EXPECTED_JSON = ('{"module_names": {"Foo": "vFoo"}, '
                     '"instance_names": '
                     '{'
                     '"Foo,bar": ["__SYMBOL_TABLE_EMPTY__", "vbar"], '
                     '"Foo,tbi": ["__SYMBOL_TABLE_INLINED_INSTANCE__", ""]'
                     '}, '
                     '"port_names": {"Foo,I": "vI"}, '
                     '"inlined_instance_names": '
                     '{'
                     '"Foo,tbi,leaf": ["__SYMBOL_TABLE_EMPTY__", "tbi_leaf"]'
                     '}, '
                     '"instance_types": {"Foo,bar": "Bar"}}')

    table = SymbolTable()
    table.set_module_name("Foo", "vFoo")
    table.set_instance_name("Foo", "bar", (SYMBOL_TABLE_EMPTY, "vbar"))
    table.set_port_name("Foo", "I", "vI")
    table.set_instance_name("Foo", "tbi", (SYMBOL_TABLE_INLINED_INSTANCE, ""))
    table.set_inlined_instance_name(
        "Foo", "tbi", "leaf", (SYMBOL_TABLE_EMPTY, "tbi_leaf"))
    table.set_instance_type("Foo", "bar", "Bar")

    as_json = table.as_json()
    assert as_json == EXPECTED_JSON

    copy = SymbolTable.from_json(as_json)
    assert table.get_module_name("Foo") == "vFoo"
    assert table.get_instance_name("Foo", "bar") == (SYMBOL_TABLE_EMPTY, "vbar")
    assert table.get_port_name("Foo", "I") == "vI"
    assert (table.get_instance_name("Foo", "tbi") ==
            (SYMBOL_TABLE_INLINED_INSTANCE, ""))
    assert (table.get_inlined_instance_name("Foo", "tbi", "leaf") ==
            (SYMBOL_TABLE_EMPTY, "tbi_leaf"))
    assert table.get_instance_type("Foo", "bar") == "Bar"
    assert copy.as_json() == EXPECTED_JSON



from magma.passes import IRPass
from magma import Bit, Circuit, IO, In, Out, wire


class _Cell(Circuit):
    io = IO(I=In(Bit), O=Out(Bit))

    io.O <= io.I


class _Top(Circuit):
    io = IO(I=In(Bit), O=Out(Bit))

    in_ = io.I
    for _ in range(5):
        cell = _Cell()
        cell.I <= in_
        in_ = cell.O
    io.O <= in_


def test_basic():
    pass_ = IRPass(_Top)
    pass_.run()
    expected = """_Cell = DefineCircuit("_Cell", "I", In(Bit), "O", Out(Bit))
wire(_Cell.I, _Cell.O)
EndCircuit()

_Top = DefineCircuit("_Top", "I", In(Bit), "O", Out(Bit))
_Cell_inst0 = _Cell()
_Cell_inst1 = _Cell()
_Cell_inst2 = _Cell()
_Cell_inst3 = _Cell()
_Cell_inst4 = _Cell()
wire(_Top.I, _Cell_inst0.I)
wire(_Cell_inst0.O, _Cell_inst1.I)
wire(_Cell_inst1.O, _Cell_inst2.I)
wire(_Cell_inst2.O, _Cell_inst3.I)
wire(_Cell_inst3.O, _Cell_inst4.I)
wire(_Cell_inst4.O, _Top.O)
EndCircuit()

"""
    assert pass_.code == expected



import os
import pathlib
import pytest
import uinspect

import magma as m
from magma.debug import get_debug_info, debug_info, magma_helper_function
from magma.passes.when import run_when_passes
from magma.backend.coreir.insert_coreir_wires import insert_coreir_wires


def test_magma_debug_ext():
    filedir = os.path.realpath(os.path.dirname(__file__))
    assert get_debug_info(2) == debug_info(
        f"{filedir}/test_debug.py",
        14,
        None
    )


@pytest.mark.parametrize(
    'op',
    (
        lambda: ~m.Bit(),
        lambda: m.Bit() | m.Bit(),
        lambda: m.Bit() == m.Bit(),
        lambda: ~m.Bits[8](),
        lambda: m.Bits[8]() | m.Bits[8](),
        lambda: m.Bits[8]() == m.Bits[8](),
        lambda: m.mux([m.Bit(), m.Bit()], m.Bit()),
        lambda: m.register(m.Bit()),
    )
)
def test_primitive_debug_info(op):

    class Foo(m.Circuit):
        op()  # just to instance some primitive

    inst = Foo.instances[0]
    filename, lineno = uinspect.get_location()
    assert inst.debug_info.filename == filename


def test_helper_function():

    class Foo(m.Circuit):
        io = m.IO()

    def _make_foo():  # no helper fn annotation
        return Foo()

    @magma_helper_function
    def _make_foo_annotated():
        return Foo()

    class Bar(m.Circuit):
        io = m.IO()
        Foo()
        _make_foo_annotated()
        _make_foo()

    instances = Bar.instances
    filename, lineno = uinspect.get_location()
    assert all(inst.debug_info.filename == filename for inst in instances)
    assert instances[0].debug_info.lineno == lineno - 5
    assert instances[1].debug_info.lineno == lineno - 4
    assert instances[2].debug_info.lineno == lineno - 13


def test_circuit_builder():

    class Builder(m.CircuitBuilder):
        def __init__(self, name):
            super().__init__(name)
            self._add_port("I", m.In(m.Bit))
            self._add_port("O", m.Out(m.Bit))
            m.wire(self._port("I"), self._port("O"))

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        b0 = Builder(name="b0")
        b0.I @= io.I
        io.O @= b0.O

    filename, lineno = uinspect.get_location()
    assert Top.instances[0].debug_info.filename == filename
    assert Top.instances[0].debug_info.lineno == lineno - 4
    assert Top.instances[0].debug_info == Top.b0.debug_info


def test_when():

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), S=m.In(m.Bit), O=m.Out(m.Bit))
        x = m.Bit(name="x")
        x @= io.I
        with m.when(io.S):
            io.O @= x
        with m.otherwise():
            io.O @= 0

    run_when_passes(Top)
    insert_coreir_wires(Top)

    filename, lineno = uinspect.get_location()
    assert Top.instances[0].debug_info.filename == filename
    assert Top.instances[0].debug_info.lineno == lineno - 8
    assert Top.instances[1].debug_info.filename == filename
    assert Top.instances[1].debug_info.lineno == lineno - 8


def test_wire():

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        x = m.Bit(name="x")
        x @= io.I
        io.O @= x

    insert_coreir_wires(Top)

    filename, lineno = uinspect.get_location()
    assert Top.instances[0].debug_info.filename == filename
    assert Top.instances[0].debug_info.lineno == lineno - 4


def test_generator_instance():

    class MyGen(m.Generator2):
        def __init__(self, w: int):
            self.io = io = m.IO(I=m.In(m.Bits[w]), O=m.Out(m.Bits[w]))
            io.O @= io.I

    MyGen4 = MyGen(4)
    MyGen8 = MyGen(8)

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bits[12]), O=m.Out(m.Bits[12]))
        x = MyGen4()(io.I[:4])
        y = MyGen8()(io.I[4:])
        io.O @= m.concat(x, y)

    filename, lineno = uinspect.get_location()
    instances = Top.instances
    assert all(inst.debug_info.filename == filename for inst in instances)
    assert instances[0].debug_info.lineno == lineno - 4
    assert instances[1].debug_info.lineno == lineno - 3



import abc
from typing import Any

from magma.backend.mlir.mlir import MlirOp


class MlirOpPass(abc.ABC):
    def __init__(self, root: MlirOp):
        self._root = root
        self._callable = callable(self)

    def _run_on_op(self, op: MlirOp):
        for region in op.regions:
            for block in region.blocks:
                for op in block.operations:
                    yield from self._run_on_op(op)
                    yield self(op)

    @abc.abstractmethod
    def __call__(self, op: MlirOp) -> Any:
        raise NotImplementedError()

    def run(self):
        yield from self._run_on_op(self._root)


class CollectMlirOpsPass(MlirOpPass):
    def __call__(self, op: MlirOp):
        return op



import pytest

import magma as m
from magma.symbol_table import SYMBOL_TABLE_EMPTY


def _compile(name, ckt, include_master=False):
    res = m.compile(name, ckt, generate_symbols=True)
    symbol_table = res["symbol_table"]
    if not include_master:
        return res["symbol_table"]
    return res["symbol_table"], res["master_symbol_table"]


class SB_DFF(m.Circuit):
    io = m.IO(D=m.In(m.Bit), Q=m.Out(m.Bit), C=m.In(m.Clock))


def test_symbol_table_dff():

    class DFFInit1(m.Circuit):
        io = m.IO(D=m.In(m.Bit), Q=m.Out(m.Bit), C=m.In(m.Clock))
        dff_inst = SB_DFF()
        dff_inst.D @= ~io.D
        io.Q @= ~dff_inst.Q

    symbol_table = _compile("build/DFFInit1", DFFInit1)
    assert symbol_table.get_module_name("DFFInit1") == "DFFInit1"
    assert (symbol_table.get_instance_name("DFFInit1", "SB_DFF_inst0") ==
            (SYMBOL_TABLE_EMPTY, "SB_DFF_inst0"))
    assert symbol_table.get_port_name("DFFInit1", "D") == "D"
    assert symbol_table.get_port_name("DFFInit1", "Q") == "Q"
    assert symbol_table.get_port_name("DFFInit1", "C") == "C"
    instance_type = symbol_table.get_instance_type("DFFInit1", "SB_DFF_inst0")
    assert instance_type == "SB_DFF"


def test_symbol_table_dff_list():

    class DFFList(m.Circuit):
        io = m.IO(D=m.In(m.Bit), Q=m.Out(m.Bit), C=m.In(m.Clock))
        dffs = []
        prev = io.D
        for i in range(10):
            dff = SB_DFF()
            dffs.append(dff)
            dff.D @= prev
            prev = dff.Q
        io.Q @= prev

    symbol_table = _compile("build/DFFList", DFFList)
    for i in range(10):
        name = symbol_table.get_instance_name("DFFList", f"SB_DFF_inst{i}")
        assert name == (SYMBOL_TABLE_EMPTY, f"SB_DFF_inst{i}")
        instance_type = symbol_table.get_instance_type(
            "DFFList", f"SB_DFF_inst{i}")
        assert instance_type == "SB_DFF"


def test_symbol_table_bundle_flattening():

    class MyBundle(m.Product):
        x = m.Bit
        y = m.Bit

    class WithTuple(m.Circuit):
        io = m.IO(
            I1=m.In(MyBundle),
            O=m.Out(MyBundle),
            I2=m.In(m.Bit)
        )
        io.O.undriven()

    symbol_table, master_symbol_table = _compile(
        "build/WithTuple", WithTuple, include_master=True)

    # Check magma -> CoreIR mapping.
    assert symbol_table.get_module_name("WithTuple") == "WithTuple"
    assert symbol_table.get_port_name("WithTuple", "I1") == "I1"
    assert symbol_table.get_port_name("WithTuple", "O") == "O"
    assert symbol_table.get_port_name("WithTuple", "I2") == "I2"
    # Check that "I.x" is not mapped at this level.
    with pytest.raises(Exception):
        symbol_table.get_port_name("WithTuple", "I1.x")

    # Check magma -> Verilog (master) mapping.
    assert master_symbol_table.get_port_name("WithTuple", "I1.x") == "I1_x"
    assert master_symbol_table.get_port_name("WithTuple", "I1.y") == "I1_y"
    assert master_symbol_table.get_port_name("WithTuple", "O.x") == "O_x"
    assert master_symbol_table.get_port_name("WithTuple", "O.y") == "O_y"
    # Check that top-level port "I1" is not mapped in the master mapping.
    with pytest.raises(Exception):
        master_symbol_table.get_port_name("WithTuple", "I1")
    assert master_symbol_table.get_port_name("WithTuple", "I2") == "I2"


def test_symbol_table_inlining_example():

    class X(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))

    class Bar(m.Circuit):
        _INLINED_ = True
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        x = X(name="x")
        io.O @= x(io.I)

    class Baz(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        bar = Bar(name="bar")
        io.O @= bar(io.I)

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        bar = Bar(name="bar")
        io.O @= bar(io.I)

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        foo = Foo(name="foo")
        baz = Baz(name="baz")
        io.O @= foo(baz(io.I))

    symbol_table = _compile("build/Top", Top)
    # NOTE: These two conditions would only hold in the post-CoreIR
    # mapping. Since we only test the magma-to-CoreIR symbol table for now,
    # these will not hold.
    #
    # assert (symbol_table.get_instance_name("Foo", "bar") ==
    #         m.symbol_table.INLINED)
    # assert symbol_table.get_inlined_instance_name("Foo", "bar", "x") == "bar_x"


def test_symbol_table_nested_inlining():

    class X(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))

    class Bar(m.Circuit):
        _INLINED_ = True
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        x = X(name="x")
        io.O @= x(io.I)

    class Baz(m.Circuit):
        _INLINED_ = True
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        bar = Bar(name="bar")
        io.O @= bar(io.I)

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        bar = Bar(name="bar")
        io.O @= bar(io.I)

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        foo = Foo(name="foo")
        baz = Baz(name="baz")
        io.O @= foo(baz(io.I))

    symbol_table = _compile("build/Top", Top)
    # NOTE: These two conditions would only hold in the post-CoreIR
    # mapping. Since we only test the magma-to-CoreIR symbol table for now,
    # these will not hold.
    #
    # assert symbol_table.get_instance_name("Top", "baz") == m.symbol_table.INLINED
    # assert symbol_table.get_instance_name("Foo", "bar") == m.symbol_table.INLINED
    # assert symbol_table.get_inlined_instance_name("Top", "baz", "bar", "x") == "baz_bar_x"



class MlirCompilerErrorBase(RuntimeError):
    pass


class MlirCompilerInternalError(MlirCompilerErrorBase):
    pass


class MlirCompilerError(MlirCompilerErrorBase):
    pass


class MlirWhenCycleError(MlirCompilerError):
    pass



from magma.common import prologue, epilogue


def test_prologue():

    def prologue_fn(lst):
        lst[0] = True

    @prologue(prologue_fn)
    def fn(lst):
        assert lst[0] is True
        lst[1] = True

    fn([False for _ in range(10)])


def test_epilogue():

    def epilogue_fn(lst):
        lst[0] = True

    @epilogue(epilogue_fn)
    def fn(lst):
        assert lst[0] is False
        lst[1] = True

    my_lst = [False for _ in range(10)]
    fn(my_lst)
    assert my_lst[0] is True



import magma as m
from magma.view import PortView


def test_port_view_ref():
    T = m.Product.from_fields("anon", dict(x=m.Bit, y=m.Bit))

    class Foo(m.Circuit):
        io = m.IO(I=m.In(T))

    x = PortView[type(Foo.I)](Foo.I, None)
    assert x._get_magma_value_().name.view is x

    y = T()
    y @= x

    assert y.value() is x._get_magma_value_()
    assert y.x.value() is x._get_magma_value_().x
    #assert y.x.value() is x.x._get_magma_value_()



import magma as m
from magma.type_utils import (isprotocol, isdigital, isbits, isarray, istuple,
                              isuint, issint, TypeVisitor, TypeTransformer)


class _Prod(m.Product):
    x = m.Array[5, m.Bits[10]]
    y = m.Array[10, m.Bit]


class _FooTypeMeta(m.MagmaProtocolMeta):
    def _to_magma_(cls):
        return cls.T

    def _qualify_magma_(cls, direction: m.Direction):
        return cls[cls.T.qualify(direction)]

    def _flip_magma_(cls):
        return cls[cls.T.flip()]

    def _from_magma_value_(cls, val: m.Type):
        return cls(val)

    def __getitem__(cls, T):
        return type(cls)(f"Foo{T}", (cls, ), {"T": T})


class _FooType(m.MagmaProtocol, metaclass=_FooTypeMeta):
    def __init__(self, val = None):
        if val is None:
            val = self.T()
        self._val = val

    def _get_magma_value_(self):
        return self._val


def test_is_functions():

    assert isprotocol(_FooType)
    assert not isprotocol(m.Bit)

    assert isdigital(m.Bit)
    assert isdigital(m.Clock)
    assert isdigital(m.Reset)
    assert not isdigital(m.Bits)

    assert isbits(m.Bits)
    assert isbits(m.Bits[8])
    assert isbits(m.UInt[8])
    assert not isbits(m.Bit)

    assert isarray(m.Array)
    assert isarray(m.Array[10, m.Bit])
    assert isarray(m.Bits)
    assert not isarray(m.Bit)

    assert istuple(_Prod)
    assert istuple(m.Product.from_fields("anon", dict(x=m.Bit)))
    assert not istuple(m.Bit)

    assert isuint(m.UInt)
    assert isuint(m.UInt[8])
    assert not isuint(m.SInt)

    assert issint(m.SInt)
    assert issint(m.SInt[8])
    assert not issint(m.UInt)


def test_type_visitor_basic():

    class _Visitor(TypeVisitor):
        def __init__(self):
            self.indent = 0
            self.string = ""

        def _print(self, s):
            self.string += s + "\n"

        def generic_visit(self, T):
            self._print(((4 * self.indent) * " ") + str(T))
            self.indent += 1
            super().generic_visit(T)
            self.indent -= 1

    T = m.Array[10, _Prod]

    v = _Visitor()
    v.visit(T)

    assert v.string == \
"""Array[(10, _Prod)]
    Tuple(x=Array[(5, Bits[10])],y=Array[(10, Bit)])
        Array[(5, Bits[10])]
            Bits[10]
        Array[(10, Bit)]
            Bit
"""


def test_type_transformer():

    class _Transformer(TypeTransformer):
        def visit_Digital(self, T):
            return m.Clock

    T = m.Array[10, _Prod]
    Tnew = _Transformer().visit(T)

    assert issubclass(Tnew, m.Array)
    assert Tnew.N == 10
    Prod = Tnew.T
    assert issubclass(Prod, m.Tuple)
    fields = Prod.field_dict
    assert set(fields.keys()) == {"x", "y"}
    assert fields["x"] is m.Array[5, m.Bits[10]]
    assert fields["y"] is m.Array[10, m.Clock]



import abc
import collections
import collections.abc
import dataclasses
from functools import wraps, partial, reduce
import hashlib
import operator
from typing import Any, Callable, Dict, Iterable, Optional
import warnings
import weakref


class Stack:
    """Lightweight wrapper over builtin lists to provide a stack interface"""
    def __init__(self):
        self._stack = []

    def push(self, value):
        self._stack.append(value)

    def pop(self):
        return self._stack.pop()

    def peek(self):
        return self._stack[-1]

    def peek_default(self, default: Any) -> Any:
        try:
            return self._stack[-1]
        except IndexError:
            return default

    def __iter__(self) -> Iterable[Any]:
        return iter(reversed(self._stack))

    def __bool__(self) -> bool:
        return bool(self._stack)


class _Ref(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value is other.value

    def __hash__(self):
        return id(self.value)


class _IdentitySetBase(collections.abc.MutableSet):
    def __init__(self, refs):
        self.refs = refs

    def __contains__(self, elem):
        return _Ref(elem) in self.refs

    def __iter__(self):
        return (ref.value for ref in self.refs)

    def __len__(self):
        return len(self.refs)

    def add(self, elem):
        self.refs.add(_Ref(elem))

    def discard(self, elem):
        raise NotImplementedError()

    def remove(self, elem):
        raise NotImplementedError()

    def pop(self):
        raise NotImplementedError()

    def clear(self):
        self.refs.clear()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, list(self))


class IdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        refs = set(map(_Ref, items))
        super().__init__(refs)


class OrderedIdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        # NOTE(rsetaluri): We use collections.OrderedDict to mimic an ordered
        # set, to avoid implementing a custom ordered set or import one, since
        # it is not natively supported.
        refs = map(lambda x: (x, None), map(_Ref, items))
        refs = collections.OrderedDict(refs)
        super().__init__(refs)

    def add(self, elem):
        self.refs[_Ref(elem)] = None


def deprecated(func=None, *, msg=None):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Marks a function as deprecated. @msg is an optional parameter to override
    the default warning.

    Examples:

        @deprecated
        def foo(...): ...

        @deprecated('Don't use bar!')
        def bar(...): ...
    """
    if func is None:
        return partial(deprecated, msg=msg)
    if msg is None:
        msg = f"{func.__name__} is deprecated"

    @wraps(func)
    def _wrapper(*args, **kwargs):
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return _wrapper


def setattrs(obj, dct, pred=None):
    for k, v in dct.items():
        if pred is None or pred(k, v):
            setattr(obj, k, v)


class ParamDict(dict):
    """
    Hashable dictionary for simple key: value parameters
    """
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def is_int(value):
    try:
        int(value)
    except (TypeError, ValueError):
        return False
    return True


def _make_delegate_fn(method):

    def _fn(self, *args, **kwargs):
        return getattr(self._underlying, method)(*args, **kwargs)

    return _fn


def make_delegator_cls(base):
    methods = base.__abstractmethods__

    class _Delegator(base):
        def __init__(self, underlying, *args, **kwargs):
            self._underlying = underlying

    for method in methods:
        setattr(_Delegator, method, _make_delegate_fn(method))
    # NOTE(rsetaluri): We should be using the new abc.update_abstractmethods
    # function. See https://bugs.python.org/issue41905 and
    # https://docs.python.org/3.10/library/abc.html#abc.update_abstractmethods.
    _Delegator.__abstractmethods__ = frozenset()
    return _Delegator


class IterableException(ValueError):
    pass


class EmptyIterableException(IterableException):
    pass


class NonSingletonIterableException(IterableException):
    pass


def only(lst: Iterable):
    it = iter(lst)
    try:
        value = next(it)
    except StopIteration:
        raise EmptyIterableException() from None
    try:
        new_value = next(it)
    except StopIteration:
        return value
    else:
        elements = [value, new_value] + list(it)
        raise NonSingletonIterableException(elements)


def is_empty(lst: Iterable) -> bool:
    """Checks whether the iterable @lst has any elements. If @lst is a
    generator, then it is modified and should no longer be used.
    """
    it = iter(lst)
    try:
        next(it)
    except StopIteration:
        return True
    return False


class Finalizable(abc.ABC):
    @abc.abstractmethod
    def finalize(self):
        raise NotImplementedError()


class FinalizableDelegator(Finalizable):
    def __init__(self):
        self._children = {}

    def add_child(self, key: str, child: Finalizable):
        if key in self._children:
            raise ValueError(f"key '{key}' already present")
        self._children[key] = child

    def get_child(self, key: str) -> Finalizable:
        return self._children[key]

    def finalize(self):
        for child in self._children.values():
            child.finalize()


def lca_of_types(classes):
    # NOTE(rsetaluri): This implementation is inspired by:
    #   https://stackoverflow.com/questions/58290137/how-to-find-most-recent-common-ancestor-base-type-of-several-types-in-python
    return next(iter(reduce(
        operator.and_, (collections.Counter(cls.mro()) for cls in classes))))


def slice_opts(dct: Dict, cls: type, keep: bool = False):
    get_opt = dct.__getitem__ if keep else dct.pop
    if not dataclasses.is_dataclass(cls):
        raise TypeError("Expected dataclass")
    kwargs = {}
    for name, field in cls.__dataclass_fields__.items():
        try:
            value = get_opt(name)
        except KeyError:
            continue
        kwargs[name] = value
    return cls(**kwargs)


def filter_by_key(function: Callable[[Any], bool], dct: Dict):
    return {k: v for k, v in dct.items() if function(k)}


def wrap_with_context_manager(ctx_mgr):

    def decorator(fn):

        @wraps(fn)
        def wrapper(*args, **kwargs):
            with ctx_mgr:
                return fn(*args, **kwargs)

        return wrapper

    return decorator


def replace_all(s: str, replacement_map: Dict[str, str]) -> str:
    for old, new in replacement_map.items():
        s = s.replace(old, new)
    return s


class SimpleCounter:
    def __init__(self, init: int = 0):
        self._value = init

    def value(self) -> int:
        return self._value

    def next(self) -> int:
        value = self._value
        self._value += 1
        return value


def sort_by_value(dct: Dict[Any, Any]) -> Iterable[Any]:
    return sorted(dct.keys(), key=lambda k: dct[k])


class MroVisitor(abc.ABC):
    def get_class(self, node: Any) -> type:
        return node.__class__

    def visit(self, node: Any, *args, **kwargs):
        method = None
        for cls in self.get_class(node).__mro__:
            name = f"visit_{cls.__name__}"
            method = getattr(self, name, None)
            if method is not None:
                break
        if method is None:
            method = self.generic_visit
        return method(node, *args, **kwargs)

    @abc.abstractmethod
    def generic_visit(self, node: Any, *args, **kwargs):
        raise NotImplementedError()


def prologue(prologue_fn):

    def decorator(fn):

        @wraps(fn)
        def wrapped(*args, **kwargs):
            prologue_fn(*args, **kwargs)
            return fn(*args, **kwargs)

        return wrapped

    return decorator


def epilogue(epilogue_fn):

    def decorator(fn):

        @wraps(fn)
        def wrapped(*args, **kwargs):
            ret = fn(*args, **kwargs)
            epilogue_fn(*args, **kwargs)
            return ret

        return wrapped

    return decorator


def assert_false(*args, **kwargs):
    assert False


def is_(x: Any, y: Any):
    return x is y


def find_by_value(
        dct: Dict[Any, Any],
        value: Any,
        eq: Callable[[Any, Any], bool] = None,
) -> Iterable[Any]:
    """Performs a reverse lookup on @dct given @value, i.e. returns any key k
    for which @eq(@dct[k], @value), where @eq is an optional equality
    function. By default `is` is used.
    """
    if eq is None:
        eq = is_
    for k, v in dct.items():
        if eq(v, value):
            yield k


def hash_expr(expr: str) -> str:
    hasher = hashlib.shake_128()
    hasher.update(expr.encode())
    return hasher.hexdigest(8)


def maybe_dereference(ref: Optional[weakref.ReferenceType]) -> Optional:
    if ref is None:
        return None
    return ref()



import pytest
import magma as m


@pytest.mark.parametrize('T', [m.Bits[8], m.Tuple[m.Bit, m.Bits[8]]])
def test_aggregate_wireable_unwire(T):
    class Foo(m.Circuit):
        io = m.IO(I=m.In(T), O=m.Out(T))
        io.O @= io.I
        assert io.O.value() is io.I
        io.O.unwire(io.I)
        assert io.O.value() is None

        io.O @= io.I
        io.O[0]  # Trigger elaboration
        assert io.O.value() is io.I
        io.O.unwire(io.I)
        assert io.O.value() is None


@pytest.mark.skip("Slow test should only be run if needed")
def test_aggregate_wireable_recursion_error():
    class T(m.Product):
        x = m.Array[2, m.Bits[8]]
        y = m.Bit

    class Foo(m.Circuit):
        T = m.Array[1024, m.Bits[8]]
        io = m.IO(I=m.In(T), O=m.Out(T), S=m.In(m.Bit))
        x = io.I
        for i in range(256):
            prev = x
            x = T(name=f"asdf{i}")
            x @= prev
        io.O @= x

    repr(Foo)

    m.compile("build/Foo", Foo, output="mlir-verilog")



import functools
import os
import pytest

import fault as f

import magma as m
from magma.config import config, EnvConfig
from magma.primitives.when import InferredLatchError
from magma.testing.utils import check_gold, update_gold, SimpleMagmaProtocol
from magma.type_utils import type_to_sanitized_string
from magma.backend.mlir.errors import MlirWhenCycleError


config.register(
    test_when_check_verilog=EnvConfig(
        "TEST_WHEN_CHECK_VERILOG", False, bool
    )
)


def _get_output_type() -> str:
    if config.test_when_check_verilog:
        return "mlir-verilog"
    return "mlir"


@functools.lru_cache(maxsize=None)
def _expects_error(error_type, process_error=None):

    def _decorator(fn):

        @functools.wraps(fn)
        def _wrapper(*args, **kwargs):
            with pytest.raises(error_type) as e:
                fn(*args, **kwargs)
            if process_error is not None:
                process_error(e)

        return _wrapper

    return _decorator


def test_with_default():
    class test_when_with_default(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S):
            io.O @= io.I[0]

    m.compile("build/test_when_with_default", test_when_with_default,
              output=_get_output_type(), emit_when_assertions=True)

    if check_gold(__file__, "test_when_with_default.mlir"):
        return

    tester = f.Tester(test_when_with_default)
    tester.poke(test_when_with_default.I, 0b10)
    tester.poke(test_when_with_default.S, 0)
    tester.eval()
    tester.expect(test_when_with_default.O, 1)
    tester.poke(test_when_with_default.S, 1)
    tester.eval()
    tester.expect(test_when_with_default.O, 0)
    tester.poke(test_when_with_default.I, 0b01)
    tester.eval()
    tester.expect(test_when_with_default.O, 1)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           magma_opts={"emit_when_assertions": True})
    update_gold(__file__, "test_when_with_default.mlir")


def test_nested_with_default():
    class test_when_nested_with_default(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S[0]) as c0:
            with m.when(io.S[1]) as c1:
                io.O @= io.I[0]

    m.compile("build/test_when_nested_with_default",
              test_when_nested_with_default, output=_get_output_type(),
              emit_when_assertions=True)

    if check_gold(__file__, "test_when_nested_with_default.mlir"):
        return

    tester = f.Tester(test_when_nested_with_default)
    tester.poke(test_when_nested_with_default.I, 0b10)
    tester.poke(test_when_nested_with_default.S, 0b00)
    tester.eval()
    tester.expect(test_when_nested_with_default.O, 1)
    tester.poke(test_when_nested_with_default.S, 0b01)
    tester.eval()
    tester.expect(test_when_nested_with_default.O, 1)
    tester.poke(test_when_nested_with_default.S, 0b10)
    tester.eval()
    tester.expect(test_when_nested_with_default.O, 1)
    tester.poke(test_when_nested_with_default.S, 0b11)
    tester.eval()
    tester.expect(test_when_nested_with_default.O, 0)
    tester.poke(test_when_nested_with_default.I, 0b01)
    tester.eval()
    tester.expect(test_when_nested_with_default.O, 1)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           magma_opts={"emit_when_assertions": True})

    update_gold(__file__, "test_when_nested_with_default.mlir")


def test_override(caplog):
    class test_when_override(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S):
            io.O @= io.I[0]
        io.O @= io.I[1]
        io.I[0].unused()

    m.compile("build/test_when_override", test_when_override,
              output=_get_output_type(), emit_when_assertions=True)

    if check_gold(__file__, "test_when_override.mlir"):
        return

    tester = f.Tester(test_when_override)
    tester.poke(test_when_override.I, 0b10)
    tester.poke(test_when_override.S, 0)
    tester.eval()
    tester.expect(test_when_override.O, 1)
    tester.poke(test_when_override.S, 1)
    tester.eval()
    tester.expect(test_when_override.O, 1)
    tester.poke(test_when_override.I, 0b01)
    tester.eval()
    tester.expect(test_when_override.O, 0)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           flags=['-Wno-UNUSED'],
                           magma_opts={"emit_when_assertions": True})

    update_gold(__file__, "test_when_override.mlir")


def test_else():
    class test_when_else(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        with m.when(io.S):
            io.O @= io.I[0]
        with m.otherwise():
            io.O @= io.I[1]

    m.compile("build/test_when_else", test_when_else, output=_get_output_type(),
              emit_when_assertions=True)

    if check_gold(__file__, "test_when_else.mlir"):
        return

    tester = f.Tester(test_when_else)
    tester.poke(test_when_else.I, 0b10)
    tester.poke(test_when_else.S, 0)
    tester.eval()
    tester.expect(test_when_else.O, 1)
    tester.poke(test_when_else.S, 1)
    tester.eval()
    tester.expect(test_when_else.O, 0)
    tester.poke(test_when_else.I, 0b01)
    tester.eval()
    tester.expect(test_when_else.O, 1)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           magma_opts={"emit_when_assertions": True})

    update_gold(__file__, "test_when_else.mlir")


def test_elsewhen():
    class test_when_elsewhen(m.Circuit):
        io = m.IO(I=m.In(m.Bits[3]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            io.O @= io.I[0]
        with m.elsewhen(io.S[1]):
            io.O @= io.I[1]
        with m.otherwise():
            io.O @= io.I[2]

    m.compile("build/test_when_elsewhen", test_when_elsewhen,
              output=_get_output_type(), emit_when_assertions=True)

    if check_gold(__file__, "test_when_elsewhen.mlir"):
        return

    tester = f.Tester(test_when_elsewhen)
    tester.poke(test_when_elsewhen.I, 0b010)
    tester.poke(test_when_elsewhen.S, 0b00)
    tester.eval()
    tester.expect(test_when_elsewhen.O, 0)
    tester.poke(test_when_elsewhen.S, 0b01)
    tester.eval()
    tester.expect(test_when_elsewhen.O, 0)
    tester.poke(test_when_elsewhen.S, 0b10)
    tester.eval()
    tester.expect(test_when_elsewhen.O, 1)
    tester.poke(test_when_elsewhen.S, 0b11)
    tester.eval()
    tester.expect(test_when_elsewhen.O, 0)
    tester.poke(test_when_elsewhen.I, 0b101)
    tester.poke(test_when_elsewhen.S, 0b10)
    tester.eval()
    tester.expect(test_when_elsewhen.O, 0)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           magma_opts={"emit_when_assertions": True})

    update_gold(__file__, "test_when_elsewhen.mlir")


def _check_err(value, name):
    expected = f"Cannot use {name} without a previous when"
    assert str(value) == expected


@pytest.mark.parametrize('fn,name', [
    (lambda x: m.elsewhen(x), 'elsewhen'),
    (lambda x: m.otherwise(), 'otherwise')])
def test_bad_syntax(fn, name):
    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[3]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        io.O @= 0

        with pytest.raises(SyntaxError) as e:
            # This should not have a _PREV_WHEN_COND to use
            with fn(io.S[1]):
                io.O @= io.I[1]
        _check_err(e.value, name)

        with m.when(io.S[0]):
            io.O @= io.I[0]

        # This when should clear _PREV_WHEN_COND
        with m.when(io.S[1]):
            # This should not have a _PREV_WHEN_COND to use
            with pytest.raises(SyntaxError) as e:
                with fn(io.S[1]):
                    io.O @= io.I[1]
            _check_err(e.value, name)


@pytest.mark.parametrize('fn,name', [
    (lambda x: m.elsewhen(x), 'elsewhen'),
    (lambda x: m.otherwise(), 'otherwise')])
def test_bad_otherwise(fn, name):
    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bits[3]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            io.O @= io.I[0]

        # This when should clear _PREV_WHEN_COND
        with m.otherwise():
            io.O @= io.I[1]

        # This should not have a _PREV_WHEN_COND to use
        with pytest.raises(SyntaxError) as e:
            with fn(io.S[1]):
                io.O @= io.I[1]
        _check_err(e.value, name)


def test_multiple_drivers():
    class test_when_multiple_drivers(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]),
                  O0=m.Out(m.Bit), O1=m.Out(m.Bit))

        io.O0 @= io.I[1]
        io.O1 @= io.I[0]
        with m.when(io.S[0]) as c0:
            with m.when(io.S[1]) as c1:
                io.O0 @= io.I[0]
                io.O1 @= io.I[1]

    m.compile("build/test_when_multiple_drivers", test_when_multiple_drivers,
              output=_get_output_type(), emit_when_assertions=True)

    if check_gold(__file__, "test_when_multiple_drivers.mlir"):
        return

    tester = f.Tester(test_when_multiple_drivers)
    tester.poke(test_when_multiple_drivers.I, 0b10)
    tester.poke(test_when_multiple_drivers.S, 0b00)
    tester.eval()
    tester.expect(test_when_multiple_drivers.O0, 1)
    tester.expect(test_when_multiple_drivers.O1, 0)
    tester.poke(test_when_multiple_drivers.S, 0b01)
    tester.eval()
    tester.expect(test_when_multiple_drivers.O0, 1)
    tester.expect(test_when_multiple_drivers.O1, 0)
    tester.poke(test_when_multiple_drivers.S, 0b10)
    tester.eval()
    tester.expect(test_when_multiple_drivers.O0, 1)
    tester.expect(test_when_multiple_drivers.O1, 0)
    tester.poke(test_when_multiple_drivers.S, 0b11)
    tester.eval()
    tester.expect(test_when_multiple_drivers.O0, 0)
    tester.expect(test_when_multiple_drivers.O1, 1)
    tester.poke(test_when_multiple_drivers.I, 0b01)
    tester.eval()
    tester.expect(test_when_multiple_drivers.O0, 1)
    tester.expect(test_when_multiple_drivers.O1, 0)

    tester.compile_and_run("verilator", magma_output="mlir-verilog",
                           directory=os.path.join(os.path.dirname(__file__),
                                                  "build"),
                           magma_opts={"emit_when_assertions": True})

    update_gold(__file__, "test_when_multiple_drivers.mlir")


@pytest.mark.parametrize('T, bits_to_fault_value, flatten_all_tuples', [
    (m.Bits[8], lambda x: x, True),
    (
        m.AnonProduct[{"x": m.Bit, "y": m.Bits[7]}],
        lambda x: (bool(x[7]), int(x[:7])),
        False
    ),
    (
        m.AnonProduct[{"x": m.Bit, "y": m.Bits[7]}],
        lambda x: (bool(x[7]), int(x[:7])),
        True
    ),
])
def test_memory(T, bits_to_fault_value, flatten_all_tuples):
    T_str = type_to_sanitized_string(T)

    class test_when_memory(m.Circuit):
        name = f"test_when_memory_{T_str}_{flatten_all_tuples}"
        io = m.IO(
            data0=m.In(T), addr0=m.In(m.Bits[5]), en0=m.In(m.Bit),
            data1=m.In(T), addr1=m.In(m.Bits[5]), en1=m.In(m.Bit),
            out=m.Out(T)
        ) + m.ClockIO()

        mem = m.Memory(32, T)()
        with m.when(io.en0):
            mem[io.addr0] @= io.data0
            io.out @= mem[io.addr1]
        with m.elsewhen(io.en1):
            mem[io.addr1] @= io.data1
            io.out @= mem[io.addr0]
        with m.otherwise():
            io.out @= m.from_bits(T, m.Bits[8](0xFF))

    m.compile(
        f"build/test_when_memory_{T_str}_{flatten_all_tuples}",
        test_when_memory,
        flatten_all_tuples=flatten_all_tuples,
        emit_when_assertions=True,
        output=_get_output_type()
    )

    if not flatten_all_tuples:
        # TODO(leonardt): fault does not support unflattened tuples
        assert check_gold(
            __file__, f"test_when_memory_{T_str}_{flatten_all_tuples}.mlir"
        )
        return

    if check_gold(
            __file__, f"test_when_memory_{T_str}_{flatten_all_tuples}.mlir"
    ):
        return

    tester = f.SynchronousTester(test_when_memory, clock=test_when_memory.CLK)
    tester.advance_cycle()
    tester.expect(test_when_memory.out, bits_to_fault_value(m.Bits[8](0xFF)))

    tester.advance_cycle()

    tester.poke(test_when_memory.data0, bits_to_fault_value(m.Bits[8](0xDE)))
    tester.poke(test_when_memory.addr0, 0xAD)

    tester.poke(test_when_memory.data1, bits_to_fault_value(m.Bits[8](0xBE)))
    tester.poke(test_when_memory.addr1, 0xEF)
    tester.poke(test_when_memory.en0, 1)
    tester.advance_cycle()

    tester.expect(test_when_memory.out, 0)
    tester.poke(test_when_memory.en0, 0)
    tester.poke(test_when_memory.en1, 1)
    tester.advance_cycle()

    tester.expect(test_when_memory.out, bits_to_fault_value(m.Bits[8](0xDE)))
    tester.poke(test_when_memory.en0, 1)
    tester.poke(test_when_memory.en1, 0)
    tester.poke(test_when_memory.addr1, 0xAD)
    tester.advance_cycle()

    tester.expect(test_when_memory.out, bits_to_fault_value(m.Bits[8](0xDE)))

    tester.compile_and_run(
        "verilator", magma_output="mlir-verilog",
        directory=os.path.join(os.path.dirname(__file__), "build"),
        magma_opts={
            "flatten_all_tuples": flatten_all_tuples,
            "emit_when_assertions": True
        },
        flags=['-Wno-UNUSED']
    )

    update_gold(__file__, f"test_when_memory_{T_str}_{flatten_all_tuples}.mlir")


@pytest.mark.parametrize(
    'T, x',
    [
        (m.Array[2, m.AnonProduct[{"x": m.Bit, "y": m.Bits[2]}]],
         [[(0, 1), (1, 0)], [(1, 0), (0, 1)]]),
        (m.AnonProduct[{"x": m.Bits[2], "y": m.Bit}],
         [(0b10, 0b1), (0b01, 0b0)])
    ]
)
@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_nested(T, x, flatten_all_tuples):
    T_str = type_to_sanitized_string(T)

    class test_when_nested(m.Circuit):
        name = f"test_when_nested_{T_str}_{flatten_all_tuples}"
        io = m.IO(I=m.In(m.Array[2, T]),
                  S=m.In(m.Bit),
                  O=m.Out(T))

        io.O @= io.I[1]
        with m.when(io.S):
            io.O @= io.I[0]

    m.compile(
        f"build/test_when_nested_{T_str}_{flatten_all_tuples}",
        test_when_nested,
        output=_get_output_type(),
        flatten_all_tuples=flatten_all_tuples,
        emit_when_assertions=True
    )

    if not flatten_all_tuples:
        # TODO(leonardt): fault does not support unflattened tuples.
        assert check_gold(
            __file__, f"test_when_nested_{T_str}_{flatten_all_tuples}.mlir"
        )
        return

    if check_gold(
        __file__, f"test_when_nested_{T_str}_{flatten_all_tuples}.mlir"
    ):
        return

    tester = f.Tester(test_when_nested)
    tester.poke(test_when_nested.I, x)
    tester.poke(test_when_nested.S, 0)
    tester.eval()
    tester.expect(test_when_nested.O, x[1])
    tester.poke(test_when_nested.S, 1)
    tester.eval()
    tester.expect(test_when_nested.O, x[0])

    tester.compile_and_run(
        "verilator",
        magma_output="mlir-verilog",
        directory=os.path.join(os.path.dirname(__file__), "build"),
        magma_opts={
            "flatten_all_tuples": flatten_all_tuples,
            "emit_when_assertions": True
        }
    )
    update_gold(__file__, f"test_when_nested_{T_str}_{flatten_all_tuples}.mlir")


def test_non_port():

    class _Test(m.Circuit):
        name = "test_non_port"
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        x = m.Bit()
        with m.when(io.S):
            x @= io.I[0]
        with m.otherwise():
            x @= io.I[1]
        io.O @= x

    basename = "test_when_non_port"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


def test_recursive_non_port():

    class _Test(m.Circuit):
        name = "test_recursive_non_port"
        io = m.IO(
            I=m.In(m.Bits[2]),
            S=m.In(m.Bit),
            O0=m.Out(m.Bit),
            O1=m.Out(m.Bit),
        )

        x = m.Bit()
        with m.when(io.S):
            x @= io.I[0]
            io.O1 @= x
        with m.otherwise():
            x @= io.I[1]
            io.O1 @= x
        io.O0 @= x

    basename = "test_when_recursive_non_port"
    m.compile(f"build/{basename}", _Test, output="mlir")
    assert check_gold(__file__, f"{basename}.mlir")


def test_internal_instantiation():

    class _Test(m.Circuit):
        name = "test_internal_instantiation"
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        with m.when(io.S):
            io.O @= io.I[0] & 1
        with m.otherwise():
            io.O @= io.I[1] ^ 1

    basename = "test_when_internal_instantiation"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


@pytest.mark.skip("Test does not pass due to when issue (#1156)")
def test_internal_instantiation_complex():

    class _Test(m.Circuit):
        name = "test_internal_instantiation_complex"
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            x = m.Bit(name="x")
            with m.when(io.S[1]):
                x @= io.I[0] & 1
            with m.otherwise():
                x @= io.I[1] ^ 1
            io.O @= x
        with m.otherwise():
            io.O @= 0b11

    basename = "test_when_internal_instantiation_complex"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


def test_register_default():

    class _Test(m.Circuit):
        name = "test_register_default"
        io = m.IO(I=m.In(m.Bit), E=m.In(m.Bit), O=m.Out(m.Bit))

        reg = m.Register(m.Bit)()

        with m.when(io.E):
            reg.I @= io.I

        io.O @= reg.O

    basename = "test_when_register_default"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


def test_register_no_default():

    class _Test(m.Circuit):
        name = "test_register_no_default"
        io = m.IO(I=m.In(m.Bit), E=m.In(m.Bit), O=m.Out(m.Bit))

        reg = m.Register(m.Bit)()

        with m.when(io.E):
            reg.I @= io.I
        with m.otherwise():
            reg.I @= ~io.I

        io.O @= reg.O

    basename = "test_when_register_no_default"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


@_expects_error(InferredLatchError)
def test_latch_error_simple():

    class _Test(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        with m.when(io.S):
            io.O @= io.I[0]

    m.compile("build/_Test", _Test, output=_get_output_type())


@_expects_error(InferredLatchError)
def test_latch_error_elsewhen():

    class _Test(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        with m.when(io.S):
            io.O @= io.I[0]
        with m.elsewhen(io.S ^ 1):
            io.O @= 1

    m.compile("build/_Test", _Test, output=_get_output_type())


@_expects_error(InferredLatchError)
def test_latch_error_nested():

    class _Test(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            with m.when(io.S[1]):
                io.O @= io.I[0]

    m.compile("build/_Test", _Test, output=_get_output_type())


def test_latch_no_error_nested():

    class _Test(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            io.O @= io.I[1]
            with m.when(io.S[1]):
                io.O @= io.I[0]
        with m.otherwise():
            io.O @= 1

    m.compile("build/_Test", _Test, output=_get_output_type())


def test_latch_no_error_nested2():

    class _Test(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            with m.when(io.S[1]):
                io.O @= io.I[0]
            with m.otherwise():
                io.O @= io.I[1]
        with m.otherwise():
            io.O @= 1

    m.compile("build/_Test", _Test, output=_get_output_type())


def test_double_elsewhen():

    class _Test(m.Circuit):
        name = "test_when_double_elsewhen"
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            io.O @= io.I[0]
        with m.elsewhen(io.S[1]):
            io.O @= io.I[1]
        with m.elsewhen(io.S.reduce_and()):
            io.O @= 1
        with m.otherwise():
            io.O @= 3

    basename = "test_when_double_elsewhen"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


def test_nested_otherwise():
    class test_when_nested_otherwise(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]),
                  O=m.Out(m.Bits[2]))

        with m.when(io.S.reduce_and()):
            io.O @= io.I
        with m.otherwise():
            with m.when(io.S[1]):
                io.O @= ~io.I
            with m.otherwise():
                io.O @= io.I

    m.compile(
        "build/test_when_nested_otherwise",
        test_when_nested_otherwise,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(__file__, "test_when_nested_otherwise.mlir")


def test_lazy_array_resolve(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_resolve"
        io = m.IO(I=m.In(m.SInt[2]), S=m.In(m.Bit), O=m.Out(m.SInt[2]))
        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O @= m.sint(m.uint(io.I) >> 1)

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array"
        io = m.IO(S=m.In(m.Bit), O=m.Out(m.Bits[2]))

        x = m.Bits[2](name="x")

        with m.when(io.S):
            x[0] @= 0
            x[1] @= 1
        with m.otherwise():
            x[0] @= 1
            x[1] @= 0

        io.O @= x

    basename = "test_when_lazy_array"
    m.compile(
        f"build/{basename}",
        _Test,
        output=_get_output_type(),
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{basename}.mlir")


def test_lazy_array_slice(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_slice"
        io = m.IO(S=m.In(m.Bit), O=m.Out(m.Bits[4]))

        x = m.Bits[4](name="x")

        with m.when(io.S):
            x[:2] @= 0
            x[2:] @= 1
        with m.otherwise():
            x[:2] @= 1
            x[2:] @= 0

        io.O @= x

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_nested(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_nested"

        T = m.Product.from_fields("anon", {"x": m.Bit, "y": m.Bit})
        io = m.IO(S=m.In(m.Bit), O=m.Out(m.Array[2, T]))

        x = m.Array[2, T](name="x")

        with m.when(io.S):
            x[0][0] @= 0
            x[0][1] @= 1
            x[1][0] @= 0
            x[1][1] @= 1
        with m.otherwise():
            x[0][0] @= 1
            x[0][1] @= 0
            x[1][0] @= 1
            x[1][1] @= 0

        io.O @= x

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_protocol(caplog):

    T = SimpleMagmaProtocol[m.Bit]

    class _Test(m.Circuit):
        name = "test_when_lazy_array_protocol"
        io = m.IO(S=m.In(m.Bit), O=m.Out(m.Array[2, T]))

        x = m.Array[2, T](name="x")

        with m.when(io.S):
            x[0] @= 0
            x[1] @= 0
        with m.otherwise():
            x[0] @= 1
            x[1] @= 1

        io.O @= x

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_slice_driving_resolve(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_slice_driving_resolve"
        io = m.IO(S=m.In(m.Bit), O=m.Out(m.Bits[4]))

        x = m.Bits[4](name="x")

        with m.when(io.S):
            x @= 2
        with m.otherwise():
            x @= 4

        io.O @= x
        x.value()[0]  # triggers driving bulk wire logic
        x[0] @= 1

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_slice_driving_resolve_2(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_slice_driving_resolve_2"
        io = m.IO(I=m.In(m.Bits[4]), S=m.In(m.Bit), O=m.Out(m.Bits[4]))

        x = m.Bits[4](name="x")

        with m.when(io.S):
            x @= io.I
        with m.otherwise():
            x @= 4

        io.O @= x
        io.I[0]  # triggers driving bulk wire logic
        x[-1] @= io.I[0]  # triggers driving bulk wire logic

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_slice_overlap(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_slice_overlap"
        io = m.IO(I=m.In(m.Bits[4]), S=m.In(m.Bit), O=m.Out(m.Bits[4]))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[:2] @= io.I[2:]
            io.O[2:] @= io.I[:2]

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_lazy_array_multiple_whens(caplog):

    class _Test(m.Circuit):
        name = "test_when_lazy_array_multiple_whens"
        io = m.IO(I=m.In(m.Bits[4]), S=m.In(m.Bit), O=m.Out(m.Bits[4]))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[:2] @= io.I[2:]
            io.O[2:] @= io.I[:2]

        with m.when(~io.S):
            io.O @= io.I

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce():

    class _Test(m.Circuit):
        name = "test_when_reg_ce"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  CE=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        with m.when(io.CE):
            x.I @= io.I
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_multiple():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_multiple"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  CE=m.In(m.Bits[2]))

        x = m.Register(m.Bits[8], has_enable=True)()
        with m.when(io.CE[0]):
            x.I @= io.I
        with m.elsewhen(io.CE[1]):
            x.I @= ~io.I

        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_when_latch_nested_multiple():
    # Should not raise a LatchError
    class _Test(m.Circuit):
        name = "test_when_latch_nested_multiple"
        io = m.IO(op=m.In(m.Bits[32]), O=m.Out(m.Bits[32]))

        pc = m.Register(m.Bits[32])()
        rc = m.Bits[32](name="rc")
        ra = m.Bits[32](0)
        rb = m.Bits[32](0)
        rai = m.Bits[8](0)
        rbi = m.Bits[8](0)
        rci = m.Bits[8](0)

        rc @= m.Bits[32](0)
        io.O @= rc
        with m.when(io.op[0]):
            pc.I @= pc.O
        with m.elsewhen(io.op[1]):
            pc.I @= 0
        with m.otherwise():
            with m.when(io.op == 0):
                rc @= ra + rb
            with m.elsewhen(io.op == 1):
                rc @= m.zext((rai << 8) | rbi, 24)
            with m.when(rci == 255):
                io.O @= 0xDE
            pc.I @= pc.O + 1


def test_reg_ce_already_wired():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_already_wired"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        x.CE @= io.x
        with m.when(io.y):
            x.I @= io.I
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_explicit_wire():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_explicit_wire"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        with m.when(io.y):
            x.I @= io.I
            x.CE @= io.x
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_explicit_wire_with_default():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_explicit_wire_with_default"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        x.CE @= 1
        with m.when(io.y):
            x.I @= io.I
            x.CE @= io.x
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_implicit_wire_twice():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_implicit_wire_twice"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        with m.when(io.y):
            x.I @= io.I
        with m.when(io.x):
            x.I @= ~io.I
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_explicit_wire_twice():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_explicit_wire_twice"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        x.CE @= ~io.y
        with m.when(io.y):
            x.I @= io.I
        with m.when(io.x):
            x.I @= ~io.I
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_user_reg():

    class Register(m.AbstractRegister):
        def __init__(self, T):
            self.io = m.IO(I=m.In(T), O=m.Out(T)) + m.ClockIO()
            self.verilog = """\
reg [7:0] x;
always @(posedge CLK)
    x <= I;
assign O = x;
"""

        def get_enable(self, inst):
            return None

    class _Test(m.Circuit):
        name = "test_when_user_reg"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]), x=m.In(m.Bit))

        x = Register(m.Bits[8])()
        with m.when(io.x):
            x.I @= io.I

        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_user_reg_enable():

    class Register(m.AbstractRegister):
        def __init__(self, T):
            self.io = m.IO(I=m.In(T), O=m.Out(T), CE=m.In(m.Enable))
            self.io += m.ClockIO()
            self.verilog = """\
    reg [7:0] x;
    always @(posedge CLK)
        if (CE)
            x <= I;
    assign O = x;
    """

        def get_enable(self, inst):
            return inst.CE

    class _Test(m.Circuit):
        name = "test_when_user_reg_enable"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]), x=m.In(m.Bits[2]))

        x = Register(m.Bits[8])()
        io.I[0]  # force elaboration to test enable logic
        with m.when(io.x[0]):
            x.I @= io.I
        with m.elsewhen(io.x[1]):
            x.I @= ~io.I

        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_when_output_resolve():

    class _Test(m.Circuit):
        name = "test_when_output_resolve"
        io = m.IO(I=m.In(m.Bits[8]),
                  x=m.In(m.Bit),
                  O0=m.Out(m.Bits[8]),
                  O1=m.Out(m.Bits[2]))

        x = m.Bits[8](name="x")
        x @= io.I

        with m.when(io.x):
            io.O0 @= x
        with m.otherwise():
            io.O0 @= ~x

        io.O1[1] @= io.O0.value()[0]  # direct reference to when builder output
        io.O1[0] @= io.O0.value()[1]  # direct reference to when builder output

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_when_output_resolve2():

    class _Test(m.Circuit):
        name = "test_when_output_resolve2"
        io = m.IO(I=m.In(m.Bits[8]),
                  x=m.In(m.Bit),
                  O0=m.Out(m.Bits[8]),
                  O1=m.Out(m.Bits[8]))

        with m.when(io.x):
            io.O0 @= io.I
        with m.otherwise():
            io.O0 @= ~io.I

        io.O1 @= io.O0.value()  # direct reference to when builder output
        io.O1[0]  # force resolve

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_when_spurious_assign(flatten_all_tuples):

    class U(m.Product):
        x = m.Bits[8]
        y = m.Bit

    class T(m.Product):
        x = m.Bits[8]
        y = U

    class _Test(m.Circuit):
        name = f"test_when_spurious_assign_{flatten_all_tuples}"
        io = m.IO(x=m.In(m.Bits[8]),
                  y=m.In(m.Bit),
                  z=m.In(m.Bits[2]),
                  O=m.Out(T))
        reg = m.Register(T, has_enable=True)()
        io.O @= reg.O

        with m.when(io.z[0]):
            with m.when(io.z[1]):
                reg.I.y.x @= io.x
            with m.otherwise():
                reg.I.y.x @= 1
        with m.otherwise():
            reg.I.y.x @= ~io.x

        with m.when(io.z[1]):
            reg.I.y.y @= io.y
        with m.otherwise():
            reg.I.y.y @= ~io.y

        reg.I.x @= io.x
        reg.CE @= io.z[0]

    m.compile(
        f"build/{_Test.name}",
        _Test,
        output=_get_output_type(),
        flatten_all_tuples=flatten_all_tuples,
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_reg_ce_implicit_override():

    class _Test(m.Circuit):
        name = "test_when_reg_ce_implicit_override"
        io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]),
                  x=m.In(m.Bit), y=m.In(m.Bit))

        x = m.Register(m.Bits[8], has_enable=True)()
        with m.when(io.y):
            x.I @= io.I
        x.CE @= io.x
        io.O @= x.O

    m.compile(f"build/{_Test.name}", _Test, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, f"{_Test.name}.mlir")


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_when_tuple_bulk_resolve(caplog, flatten_all_tuples):

    class V(m.Product):
        x = m.Bits[8]
        y = m.Bits[8]

    class U(m.Product):
        x = V
        y = V

    class T(m.Product):
        x = m.Bits[8]
        y = U

    class _Test(m.Circuit):
        name = f"test_when_tuple_bulk_resolve_{flatten_all_tuples}"
        io = m.IO(I=m.In(T), S=m.In(m.Bits[2]), O=m.Out(T))

        x = m.Register(T)(name="x")
        y = m.Register(T)(name="y")
        with m.when(io.S[0]):
            x.I @= x.O
            y.I @= y.O
        with m.otherwise():
            with m.when(io.S[1]):
                x.I.x @= io.I.x
                x.I.y @= io.I.y
            with m.otherwise():
                x.I @= x.O
            y.I @= x.O
        io.O @= x.O

    assert not caplog.messages
    m.compile(
        f"build/{_Test.name}",
        _Test,
        output=_get_output_type(),
        flatten_all_tuples=flatten_all_tuples,
        emit_when_assertions=True
    )
    assert check_gold(__file__, f"{_Test.name}.mlir")


def test_when_type_error():

    with pytest.raises(TypeError):
        class Foo(m.Circuit):
            io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
            with m.when(io.I):
                assert False, "Should raise type error for non bit type"

    with pytest.raises(TypeError):
        class Foo(m.Circuit):
            io = m.IO(I=m.In(m.Bits[8]), O=m.Out(m.Bits[8]))
            with m.when(io.I[0]):
                io.O @= 1
            with m.elsewhen(io.I):
                assert False, "Should raise type error for non bit type"


def test_when_partial_array_assign():
    class test_when_partial_array_assign(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bits[2]))

        io.O[0] @= io.I[1]
        with m.when(io.S):
            io.O[1] @= io.I[0]
        with m.otherwise():
            io.O[1] @= io.I[1]

    m.compile("build/test_when_partial_array_assign",
              test_when_partial_array_assign, output=_get_output_type(),
              emit_when_assertions=True)

    assert check_gold(__file__, "test_when_partial_array_assign.mlir")


def test_when_2d_array_assign():
    class test_when_2d_array_assign(m.Circuit):
        io = m.IO(I=m.In(m.Array[2, m.Bits[2]]), S=m.In(m.Bit),
                  O=m.Out(m.Array[2, m.Bits[2]]))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[0] @= io.I[1]
            io.O[1] @= io.I[0]

    m.compile(
        "build/test_when_2d_array_assign",
        test_when_2d_array_assign,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(__file__, "test_when_2d_array_assign.mlir")


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_when_nested_array_assign(flatten_all_tuples):
    class T(m.Product):
        x = m.Bit
        y = m.Tuple[m.Bit, m.Bits[8]]

    class test_when_nested_array_assign(m.Circuit):
        name = f"test_when_nested_array_assign_{flatten_all_tuples}"
        io = m.IO(I=m.In(T), S=m.In(m.Bit),
                  O=m.Out(T))
        io.O.x @= io.I.x
        io.O.y[0] @= io.I.y[0]

        with m.when(io.S):
            io.O.y[1] @= io.I.y[1]
        with m.otherwise():
            io.O.y[1] @= io.I.y[1][::-1]

    m.compile(
        f"build/test_when_nested_array_assign_{flatten_all_tuples}",
        test_when_nested_array_assign, output=_get_output_type(),
        flatten_all_tuples=flatten_all_tuples,
        emit_when_assertions=True
    )

    assert check_gold(
        __file__, f"test_when_nested_array_assign_{flatten_all_tuples}.mlir"
    )


def test_when_partial_assign_order():
    class test_when_partial_assign_order(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]),
                  O0=m.Out(m.Bits[2]), O1=m.Out(m.Bits[2]),
                  O2=m.Out(m.Bits[2]))

        io.O0 @= io.I
        with m.when(io.S[0]):
            io.O1 @= m.bits((~io.I)[::-1])
            io.O2 @= io.I
        with m.elsewhen(io.S[1]):
            io.O1 @= io.I
            io.O2 @= io.I
        with m.otherwise():
            io.O1 @= ~io.I
            io.O0 @= m.bits(io.I[::-1])
            io.O2 @= ~io.I

    m.compile("build/test_when_partial_assign_order",
              test_when_partial_assign_order, output=_get_output_type(),
              emit_when_assertions=True)

    assert check_gold(__file__, "test_when_partial_assign_order.mlir")


def test_when_3d_array_assign():
    class test_when_3d_array_assign(m.Circuit):
        T = m.Array[2, m.Array[2, m.Bits[2]]]
        io = m.IO(I=m.In(T), S=m.In(m.Bit), O=m.Out(T))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[1][0] @= io.I[0][1]
            io.O[1][1] @= io.I[0][0]
            io.O[0][0] @= io.I[1][1]
            io.O[0][1] @= io.I[1][0]

    m.compile(
        "build/test_when_3d_array_assign",
        test_when_3d_array_assign,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(__file__, "test_when_3d_array_assign.mlir")


def test_when_array_resolved_after():
    class test_when_array_resolved_after(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), S=m.In(m.Bit), O=m.Out(m.Bits[16]))

        x = m.Bits[8]()

        x @= ~io.I
        with m.when(io.S):
            x @= io.I

        io.O @= m.concat(m.Bits[8](0), x)  # trigger resolve

    m.compile(
        "build/test_when_array_resolved_after",
        test_when_array_resolved_after,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(__file__, "test_when_array_resolved_after.mlir")


def test_when_array_3d_bulk_child():
    class test_when_array_3d_bulk_child(m.Circuit):
        T = m.Array[2, m.Array[2, m.Bits[2]]]
        io = m.IO(I=m.In(T), S=m.In(m.Bit), O=m.Out(T))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[1] @= io.I[0]
            io.O[0] @= io.I[1]

    m.compile(
        "build/test_when_array_3d_bulk_child",
        test_when_array_3d_bulk_child,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(__file__, "test_when_array_3d_bulk_child.mlir")


def test_when_temporary_resolved():
    class test_when_temporary_resolved(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), S=m.In(m.Bits[2]), O=m.Out(m.Bits[8]))

        N = 8
        const = 4
        count = m.Register(m.Bits[N])()
        x = m.zext_to(count.O, N + 1) + m.zext_to(io.I, N + 1)
        y = x[0:N]
        z = m.zext_to((count.O - const), N)

        countNext = m.Bits[N]()

        with m.when(io.S[0]):
            countNext @= z
        with m.elsewhen(io.S[1]):
            countNext @= y
        with m.otherwise():
            countNext @= y - const

        # Force resolve
        for i in range(8):
            count.I[i] @= countNext[i]
        io.O @= countNext

    m.compile("build/test_when_temporary_resolved",
              test_when_temporary_resolved, output=_get_output_type(),
              disallow_local_variables=True, emit_when_assertions=True)

    assert check_gold(__file__, "test_when_temporary_resolved.mlir")


def test_when_2d_array_assign_slice():
    class test_when_2d_array_assign_slice(m.Circuit):
        io = m.IO(I=m.In(m.Array[4, m.Bits[2]]), S=m.In(m.Bit),
                  O=m.Out(m.Array[4, m.Bits[2]]))

        with m.when(io.S):
            io.O @= io.I
        with m.otherwise():
            io.O[:2] @= io.I[2:]
            io.O[2:] @= io.I[:2]

    m.compile("build/test_when_2d_array_assign_slice",
              test_when_2d_array_assign_slice, output=_get_output_type(),
              emit_when_assertions=True)

    assert check_gold(__file__, "test_when_2d_array_assign_slice.mlir")


def test_when_unique():

    class Gen(m.Generator2):
        def __init__(self, width: int):
            self.io = io = m.IO(
                a=m.In(m.Bits[8]),
                y=m.Out(m.Bits[8]),
            )
            with m.when(io.a[0]):
                io.y @= m.zext_to(io.a[:width], 8)
            with m.otherwise():
                io.y @= io.a

    class test_when_unique(m.Circuit):
        io = m.IO(
            a=m.In(m.Bits[8]),
            y=m.Out(m.Bits[8]),
        )
        io.y @= Gen(2)()(io.a) | Gen(4)()(io.a)

    m.compile("build/test_when_unique", test_when_unique, output=_get_output_type(),
              emit_when_assertions=True)
    assert check_gold(__file__, "test_when_unique.mlir")


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_when_tuple_as_bits_resolve(flatten_all_tuples):

    class V(m.Product):
        x = m.Bits[8]
        y = m.Bit

    class T(m.Product):
        y = m.Bit
        z = V
        x = m.Array[2, m.Bits[8]]

    class U(m.Product):
        x = T
        y = m.Bits[8]

    class test_when_tuple_as_bits_resolve(m.Circuit):
        name = f"test_when_tuple_as_bits_resolve_{flatten_all_tuples}"
        io = m.IO(I=m.In(U), S=m.In(m.Bit),
                  O=m.Out(U), X=m.Out(m.Bits[U.flat_length()]))
        io.O.y @= io.I.y
        with m.when(io.S):
            for i in range(2):
                io.O.x.x[i] @= io.I.x.x[i]
            io.O.x.y @= io.I.x.y
            for key, port in io.O.x.items():
                if key == "z":
                    port @= m.mux([io.I.x.z, io.I.x.z], io.S)
        with m.otherwise():
            io.O @= io.I
        io.X @= m.as_bits(io.O.value())

    m.compile(
        f"build/test_when_tuple_as_bits_resolve_{flatten_all_tuples}",
        test_when_tuple_as_bits_resolve,
        output=_get_output_type(),
        flatten_all_tuples=flatten_all_tuples,
        disallow_local_variables=True,
        emit_when_assertions=True
    )
    assert check_gold(
        __file__, f"test_when_tuple_as_bits_resolve_{flatten_all_tuples}.mlir"
    )


def test_when_emit_asserts_basic():

    class test_when_emit_asserts_basic(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S):
            io.O @= io.I[0]

    m.compile("build/test_when_emit_asserts_basic",
              test_when_emit_asserts_basic,
              output=_get_output_type(), emit_when_assertions=True)
    assert check_gold(__file__, "test_when_emit_asserts_basic.mlir")


def test_when_emit_asserts_tuple():

    T = m.Tuple[m.Bits[8], m.Bit]

    class test_when_emit_asserts_tuple(m.Circuit):
        io = m.IO(I=m.In(m.Array[2, T]), S=m.In(m.Bit), O=m.Out(T))

        io.O @= io.I[1]
        with m.when(io.S):
            io.O @= io.I[0]

    m.compile("build/test_when_emit_asserts_tuple",
              test_when_emit_asserts_tuple,
              flatten_all_tuples=True,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_tuple.mlir")


def test_when_emit_asserts_otherwise():

    class test_when_emit_asserts_otherwise(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bit), O=m.Out(m.Bit))

        with m.when(io.S):
            io.O @= io.I[0]
        with m.otherwise():
            io.O @= io.I[1]

    m.compile("build/test_when_emit_asserts_otherwise",
              test_when_emit_asserts_otherwise,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_otherwise.mlir")


def test_when_emit_asserts_elsewhen():

    class test_when_emit_asserts_elsewhen(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S[0]):
            io.O @= io.I[0]
        with m.elsewhen(io.S[1]):
            io.O @= ~io.I[0]

    m.compile("build/test_when_emit_asserts_elsewhen",
              test_when_emit_asserts_elsewhen,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_elsewhen.mlir")


def test_when_emit_asserts_elsewhen_otherwise():

    class test_when_emit_asserts_elsewhen_otherwise(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        with m.when(io.S[0]):
            io.O @= io.I[0]
        with m.elsewhen(io.S[1]):
            io.O @= ~io.I[0]
        with m.otherwise():
            io.O @= io.I[1]

    m.compile("build/test_when_emit_asserts_elsewhen_otherwise",
              test_when_emit_asserts_elsewhen_otherwise,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(
        __file__, "test_when_emit_asserts_elsewhen_otherwise.mlir"
    )


def test_when_emit_asserts_nesting():

    class test_when_emit_asserts_nesting(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        io.O @= io.I[1]
        with m.when(io.S[0]):
            with m.when(io.S[1]):
                io.O @= io.I[0]
            with m.elsewhen(io.S.reduce_and()):
                io.O @= ~io.I[0]

    m.compile("build/test_when_emit_asserts_nesting",
              test_when_emit_asserts_nesting,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_nesting.mlir")


def test_when_emit_asserts_chained():

    class test_when_emit_asserts_chained(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), S=m.In(m.Bits[2]), O=m.Out(m.Bit))

        x = m.Bit()

        x @= io.I[1]
        with m.when(io.S[0]):
            x @= io.I[0]

        io.O @= x
        with m.when(io.S[1]):
            io.O @= ~x

    m.compile("build/test_when_emit_asserts_chained",
              test_when_emit_asserts_chained,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_chained.mlir")


def test_when_emit_asserts_value():

    class T(m.Product):
        x = m.Bit
        y = m.Bits[8]

    class test_when_emit_asserts_value(m.Circuit):
        io = m.IO(S=m.In(m.Bits[3]), O=m.Producer(m.ReadyValid[T]))

        reg = m.Register(T)()
        io.O.valid @= 0
        with m.when(io.S[0]):
            io.O.data @= reg.O
        with m.otherwise():
            io.O.data.x @= m.mux([False, True], io.S[1])
            io.O.data.y @= m.mux([m.Bits[8](0), m.Bits[8](1)], io.S[2])

        with m.when(io.S[0]):
            with m.when(io.S[1]):
                reg.I @= io.O.data.value()
            with m.elsewhen(io.S[2]):
                reg.I @= io.O.data.value()

    m.compile("build/test_when_emit_asserts_value",
              test_when_emit_asserts_value,
              flatten_all_tuples=True,
              output=_get_output_type(), emit_when_assertions=True)

    assert check_gold(__file__, "test_when_emit_asserts_value.mlir")


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_when_emit_asserts_tuple_elab(flatten_all_tuples):

    T = m.Tuple[m.Bits[8], m.Bit]

    class test_when_emit_asserts_tuple_elab(m.Circuit):
        name = f"test_when_emit_asserts_tuple_elab_{flatten_all_tuples}"
        io = m.IO(I=m.In(m.Array[2, T]), S=m.In(m.Bit), O=m.Out(T))

        x = m.Register(T)()
        io.O @= x.O
        with m.when(io.S):
            x.I @= io.O.value()
        with m.elsewhen(io.I[0][1]):
            x.I @= io.O.value()

    m.compile(
        f"build/{test_when_emit_asserts_tuple_elab.name}",
        test_when_emit_asserts_tuple_elab,
        flatten_all_tuples=flatten_all_tuples,
        output=_get_output_type(),
        emit_when_assertions=True
    )

    assert check_gold(
        __file__, f"{test_when_emit_asserts_tuple_elab.name}.mlir"
    )


def _check_or_update(circ):
    # We check verilog here because the alwcomb order was "legal" MLIR.
    m.compile(f"build/{circ.name}", circ, output="mlir-verilog")

    _file = f"{circ.name}.v"
    if check_gold(__file__, _file):
        return
    verilator_path = os.path.join(
        os.path.dirname(__file__),
        "build",
        _file
    )
    assert not os.system(f"verilator --lint-only {verilator_path}")
    update_gold(__file__, _file)


def test_when_alwcomb_order():

    class test_when_alwcomb_order(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), S=m.In(m.Bits[1]), O=m.Out(m.Bits[8]))
        x = m.Bits[8]()

        io.O @= x
        with m.when(io.S[0]):
            x @= io.I
        with m.otherwise():
            x @= ~io.I
            io.O @= ~x

    _check_or_update(test_when_alwcomb_order)


def test_when_alwcomb_order_complex():
    class test_when_alwcomb_order_complex(m.Circuit):
        io = m.IO(I=m.In(m.Bits[8]), S=m.In(m.Bits[2]), O=m.Out(m.Bits[8]))
        x = m.Bits[8]()

        io.O @= x
        with m.when(io.S[0]):
            x @= io.I
            with m.when(io.S[1]):
                x @= ~io.I
                io.O @= x & 0xDE
        with m.elsewhen(io.S[0] ^ io.S[1]):
            x @= io.I ^ io.I
        with m.otherwise():
            io.O @= ~x
            x @= ~io.I

    _check_or_update(test_when_alwcomb_order_complex)


def test_when_alwcomb_order_nested():
    class T(m.Product):
        x = m.Bit
        y = m.Bits[8]

    class test_when_alwcomb_order_nested(m.Circuit):
        io = m.IO(I=m.In(T), S=m.In(m.Bit), O=m.Out(T))
        x = T()

        io.O @= io.I
        with m.when(io.S):
            io.O @= x
            x.x @= io.I.x
            x.y @= io.I.y
        with m.otherwise():
            x.x @= ~io.I.x
            x.y @= ~io.I.y

    _check_or_update(test_when_alwcomb_order_nested)


def test_when_alwcomb_order_nested_2():
    class T(m.Product):
        x = m.Bit
        y = m.Bits[8]

    class test_when_alwcomb_order_nested_2(m.Circuit):
        io = m.IO(I=m.In(m.Array[3, T]), S=m.In(m.Bit), O=m.Out(T))
        x = T()

        io.O @= io.I[0]
        with m.when(io.S):
            x @= io.I[1]
        with m.otherwise():
            io.O.x @= x.x
            io.O.y @= x.y
            x @= io.I[2]

    _check_or_update(test_when_alwcomb_order_nested_2)


# TODO: In this case, we'll generate elaborated assignments, but it should
# be possible for us to pack these into a concat/create assignment
# def test_when_2d_array_assign():
#     class test_when_2d_array_assign(m.Circuit):
#         io = m.IO(I=m.In(m.Array[2, m.Bits[2]]), S=m.In(m.Bit),
#                   O=m.Out(m.Array[2, m.Bits[2]]))

#         with m.when(io.S):
#             io.O[0] @= io.I[1]
#             io.O[1] @= io.I[0]
#         with m.otherwise():
#             io.O[0] @= io.I[0]
#             io.O[1] @= io.I[1]

#     m.compile("build/test_when_2d_array_assign",
#               test_when_2d_array_assign, output="mlir")

#     assert check_gold(__file__, "test_when_2d_array_assign.mlir")



import pytest
from typing import Optional

import magma as m
from magma.passes.passes import CircuitPass
import magma.testing


class _MyWrapperGen(m.Generator2):
    def __init__(self, width: Optional[int]):
        T = m.Bits[width] if width is not None else m.Bit
        self.io = io = m.IO(I=m.In(T), O=m.Out(T)) + m.ClockIO()
        m.inline_verilog(
            f"reg [{width - 1}:0] R;\nasssign R <= {{I}};\n",
            I=io.I
        )
        io.O @= m.InlineVerilogExpression("R | I", T)()()

        self.name = f"MyWrapperGen_{width}"


class _Top(m.Circuit):
    name = "Top"
    T = m.Bits[2]
    io = m.IO(I=m.In(T), O=m.Out(T)) + m.ClockIO()
    tmp = _MyWrapperGen(1)()(m.bits([io.I[0]]))
    tmp = m.bits([tmp[0]] * len(T))
    io.O @= _MyWrapperGen(2)()(tmp)


class _BuildRepr(CircuitPass):
    def __init__(self, main):
        super().__init__(main)
        self._repr_str = ""

    def __call__(self, ckt):
        self._repr_str += repr(ckt)

    @property
    def repr_str(self) -> str:
        return self._repr_str


def test_basic():
    expected = \
"""term = DeclareCircuit("term", "I", In(Bits[1]))_MyWrapperGen_inline_verilog_0 = DefineCircuit("_MyWrapperGen_inline_verilog_0", "__magma_inline_value_0", In(Bits[1]))
term_inst0 = term()
wire(_MyWrapperGen_inline_verilog_0.__magma_inline_value_0, term_inst0.in)
EndCircuit()InlineVerilogExpression_Bits1_a489b4461c0192fc = DeclareCircuit("InlineVerilogExpression_Bits1_a489b4461c0192fc", "O", Out(Bits[1]))MyWrapperGen_1 = DefineCircuit("MyWrapperGen_1", "I", In(Bits[1]), "O", Out(Bits[1]), "CLK", In(Clock))
InlineVerilogExpression_Bits1_a489b4461c0192fc_inst0 = InlineVerilogExpression_Bits1_a489b4461c0192fc()
_MyWrapperGen_inline_verilog_inst_0 = _MyWrapperGen_inline_verilog_0(name="_MyWrapperGen_inline_verilog_inst_0")
wire(MyWrapperGen_1.I, _MyWrapperGen_inline_verilog_inst_0.__magma_inline_value_0)
wire(InlineVerilogExpression_Bits1_a489b4461c0192fc_inst0.O, MyWrapperGen_1.O)
EndCircuit()term = DeclareCircuit("term", "I", In(Bits[2]))_MyWrapperGen_inline_verilog_0 = DefineCircuit("_MyWrapperGen_inline_verilog_0", "__magma_inline_value_0", In(Bits[2]))
term_inst0 = term()
wire(_MyWrapperGen_inline_verilog_0.__magma_inline_value_0, term_inst0.in)
EndCircuit()InlineVerilogExpression_Bits2_a489b4461c0192fc = DeclareCircuit("InlineVerilogExpression_Bits2_a489b4461c0192fc", "O", Out(Bits[2]))MyWrapperGen_2 = DefineCircuit("MyWrapperGen_2", "I", In(Bits[2]), "O", Out(Bits[2]), "CLK", In(Clock))
InlineVerilogExpression_Bits2_a489b4461c0192fc_inst0 = InlineVerilogExpression_Bits2_a489b4461c0192fc()
_MyWrapperGen_inline_verilog_inst_0 = _MyWrapperGen_inline_verilog_0(name="_MyWrapperGen_inline_verilog_inst_0")
wire(MyWrapperGen_2.I, _MyWrapperGen_inline_verilog_inst_0.__magma_inline_value_0)
wire(InlineVerilogExpression_Bits2_a489b4461c0192fc_inst0.O, MyWrapperGen_2.O)
EndCircuit()Top = DefineCircuit("Top", "I", In(Bits[2]), "O", Out(Bits[2]), "CLK", In(Clock))
MyWrapperGen_1_inst0 = MyWrapperGen_1()
MyWrapperGen_2_inst0 = MyWrapperGen_2()
wire(Top.I[0], MyWrapperGen_1_inst0.I[0])
wire(MyWrapperGen_1_inst0.O[0], MyWrapperGen_2_inst0.I[0])
wire(MyWrapperGen_1_inst0.O[0], MyWrapperGen_2_inst0.I[1])
wire(MyWrapperGen_2_inst0.O, Top.O)
EndCircuit()"""
    p = _BuildRepr(_Top)
    p.run()
    assert p.repr_str == expected


def test_compilation_coreir():
    with pytest.raises(NotImplementedError) as e:
        m.compile("", _Top, output="coreir")


def test_compilation_mlir():
    basename = f"test_inline_verilog_expression_compilation_mlir"
    m.compile(f"build/{basename}", _Top, output="mlir")
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.mlir", f"gold/{basename}.mlir")
